
# Built-in
import os
import warnings
import datetime as dtm      # DB

# Common
import numpy as np
import scipy.optimize as scpopt
import scipy.interpolate as scpinterp
import scipy.sparse as sparse
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt


# ToFu-specific
from ._fit12d_dinput import *
from ._fit12d_dinput import _dict2vector_dscalesx0bounds
from . import _fit12d_funccostjac as _funccostjac
from ._fit12d_dextract import *
from . import _plot


__all__ = [
    'fit1d_dinput',
    'fit2d_dinput',
    'fit12d_dvalid',
    'fit12d_dscales',
    'fit1d',
    'fit2d',
    'fit1d_extract',
    'fit2d_extract',
]


_TOL1D = {'x': 1e-10, 'f': 1.e-10, 'g': 1.e-10}
_TOL2D = {'x': 1e-6, 'f': 1.e-6, 'g': 1.e-6}
_CHAIN = True
_METHOD = 'trf'
_LOSS = 'linear'
_SIGMA_MARGIN = 3.
_DVALIDITY = {
    0: 'ok',
    -1: 'non-valid input data',
    -2: 'convergence failed',
    -3: 'insufficient lines / bck',
}



###########################################################
###########################################################
#
#           Load dinput
#
###########################################################
###########################################################


def _rebuild_dict(dd):
    for k0, v0 in dd.items():
        if isinstance(v0, np.ndarray) and v0.shape == ():
            dd[k0] = v0.tolist()
        if isinstance(dd[k0], dict):
            _rebuild_dict(dd[k0])


def _checkformat_dinput(dinput, allow_pickle=True):
    if isinstance(dinput, str):
        if not (os.path.isfile(dinput) and dinput[-4:] == '.npz'):
            msg = ("Arg dinput must be aither a dict or "
                   + "the absolute path to a .npz\n"
                   + "  You provided: {}".format(dinput))
            raise Exception(msg)
        dinput = dict(np.load(dinput, allow_pickle=allow_pickle))

    if not isinstance(dinput, dict):
        msg = (
            "dinput must be a dict!\n"
            + "  You provided: {}".format(type(dinput))
        )

    _rebuild_dict(dinput)
    return dinput


###########################################################
###########################################################
#
#           Main fitting sub-routines
#
###########################################################
###########################################################


def _checkformat_options(
    chain, method, tr_solver, tr_options,
    xtol, ftol, gtol, loss, max_nfev, verbose, strict,
):
    if chain is None:
        chain = _CHAIN
    if method is None:
        method = _METHOD
    assert method in ['trf', 'dogbox'], method
    if tr_solver is None:
        tr_solver = None
    if tr_options is None:
        tr_options = {}
    if xtol is None:
        xtol = _TOL1D['x']
    if ftol is None:
        ftol = _TOL1D['f']
    if gtol is None:
        gtol = _TOL1D['g']
    if loss is None:
        loss = _LOSS
    if max_nfev is None:
        max_nfev = None
    if verbose is None:
        verbose = 1
    if verbose == 3:
        verbscp = 2
    else:
        verbscp = 0
    if strict is None:
        strict = False

    return (chain, method, tr_solver, tr_options,
            xtol, ftol, gtol, loss, max_nfev, verbose, verbscp, strict)


def multigausfit1d_from_dlines(
    dinput=None,
    method=None, tr_solver=None, tr_options=None,
    xtol=None, ftol=None, gtol=None,
    max_nfev=None, chain=None, verbose=None,
    loss=None, jac=None,
    strict=None,
):
    """ Solve multi_gaussian fit in 1d from dlines

    If double is True, all lines are double with common shift and ratio

    Unknowns are:
        x = [bck, w0, v0, c00, c01, ..., c0n, w1, v1, c10, c11, ..., c1N, ...]

        - bck : constant background
        - wi  : spectral width of a group of lines (ion): wi^2 = 2kTi / m*c**2
                This way, it is dimensionless
        - vni : normalised velicity of the ion: vni = vi / c
        - cij : normalised coef (intensity) of line: cij = Aij

    Scaling is done so each quantity is close to unity:
        - bck: np.mean(data[data < mean(data)/2])
        - wi : Dlamb / 20
        - vni: 10 km/s
        - cij: np.mean(data)

    """

    # ---------------------------
    # Check format options
    (
        chain, method, tr_solver, tr_options,
        xtol, ftol, gtol, loss, max_nfev,
        verbose, verbscp, strict,
    ) = _checkformat_options(
         chain, method, tr_solver, tr_options,
         xtol, ftol, gtol, loss, max_nfev, verbose, strict,
    )

    # ---------------------------
    # Load dinput if necessary
    dinput = _checkformat_dinput(dinput)
    dprepare, dind = dinput['dprepare'], dinput['dind']
    nspect = dprepare['data'].shape[0]

    # ---------------------------
    # If same spectrum => consider a single data set
    if dinput['same_spectrum'] is True:
        lamb = (
            dinput['same_spectrum_dlamb']*np.arange(0, nspect)[:, None]
            + dprepare['lamb'][None, :]
        ).ravel()
        datacost = dprepare['data'].ravel()[None, :]
        nspect = data.shape[0]
        chain = False
    else:
        lamb = dprepare['lamb']
        datacost = dprepare['data']

    # ---------------------------
    # Get scaling, x0, bounds from dict
    scales = _dict2vector_dscalesx0bounds(
        dd=dinput['dscales'], dd_name='dscales', dinput=dinput,
    )
    x0 = _dict2vector_dscalesx0bounds(
        dd=dinput['dx0'], dd_name='dx0', dinput=dinput,
    )
    boundmin = _dict2vector_dscalesx0bounds(
        dd=dinput['dbounds']['min'], dd_name="dbounds['min']", dinput=dinput,
    )
    boundmax = _dict2vector_dscalesx0bounds(
        dd=dinput['dbounds']['max'], dd_name="dbounds['max']", dinput=dinput,
    )
    bounds = np.array([boundmin[0, :], boundmax[0, :]])

    # ---------------------------
    # Separate free from constant parameters
    const = _dict2vector_dscalesx0bounds(
        dd=dinput['dconstants'], dd_name='dconstants', dinput=dinput,
    )
    indx = np.any(np.isnan(const), axis=0)
    const = const[:, ~indx]
    x0[:, ~indx] = const / scales[:, ~indx]

    # ---------------------------
    # Get function, cost function and jacobian
    (
        func_detail, func_cost, func_jac,
    ) = _funccostjac.multigausfit1d_from_dlines_funccostjac(
        lamb, dinput=dinput, dind=dind, jac=jac, indx=indx,
    )

    # ---------------------------
    # Prepare output
    sol_x = np.full((nspect, dind['sizex']), np.nan)
    success = np.full((nspect,), np.nan)
    time = np.full((nspect,), np.nan)
    cost = np.full((nspect,), np.nan)
    nfev = np.full((nspect,), np.nan)
    validity = np.zeros((nspect,), dtype=int)
    message = ['' for ss in range(nspect)]
    errmsg = ['' for ss in range(nspect)]

    # Prepare msg
    if verbose in [1, 2]:
        col = np.char.array([
            'spect', 'time (s)', 'cost', 'nfev', 'njev', 'msg',
        ])
        maxl = max(np.max(np.char.str_len(col)), 10)
        msg = '\n'.join([' '.join([cc.ljust(maxl) for cc in col]),
                         ' '.join(['-'*maxl]*6)])
        print(msg)

    # ------------
    # Main loop

    end = '\r'
    t0 = dtm.datetime.now()     # DB
    for ii in range(nspect):

        if verbose == 3:
            msg = "\nspect {} / {}".format(ii+1, nspect)
            print(msg)

        try:
            dti = None
            t0i = dtm.datetime.now()     # DB
            if not dinput['valid']['indt'][ii]:
                validity[ii] = -1
                continue

            # optimization
            res = scpopt.least_squares(
                func_cost,
                x0[ii, indx],
                jac=func_jac,
                bounds=bounds[:, indx],
                method=method,
                ftol=ftol,
                xtol=xtol,
                gtol=gtol,
                x_scale=1.0,
                f_scale=1.0,
                loss=loss,
                diff_step=None,
                tr_solver=tr_solver,
                tr_options=tr_options,
                jac_sparsity=None,
                max_nfev=max_nfev,
                verbose=verbscp,
                args=(),
                kwargs={
                    'data': datacost[ii, :],
                    'scales': scales[ii, :],
                    'const': const[ii, :],
                    'indok': dprepare['indok_bool'][ii, :],
                },
            )
            dti = (dtm.datetime.now() - t0i).total_seconds()

            if chain is True and ii < nspect-1:
                x0[ii+1, indx] = res.x

            # cost, message, time
            success[ii] = res.success
            cost[ii] = res.cost
            nfev[ii] = res.nfev
            message[ii] = res.message
            time[ii] = round(
                (dtm.datetime.now()-t0i).total_seconds(),
                ndigits=3,
            )
            sol_x[ii, indx] = res.x
            sol_x[ii, ~indx] = const[ii, :] / scales[ii, ~indx]

        except Exception as err:
            if strict:
                raise err
            else:
                errmsg[ii] = str(err)
                validity[ii] = -2

        # verbose
        if verbose in [1, 2]:
            if validity[ii] == 0:
                col = np.char.array([
                    '{} / {}'.format(ii+1, nspect),
                    '{}'.format(dti),
                    '{:5.3e}'.format(res.cost),
                    str(res.nfev),
                    str(res.njev),
                    res.message,
                ])
            else:
                col = np.char.array([
                    '{} / {}'.format(ii+1, nspect),
                    '{}'.format(dti),
                    ' - ', ' - ', ' - ',
                    errmsg[ii],
                ])
            msg = ' '.join([cc.ljust(maxl) for cc in col])
            if verbose == 1:
                if ii == nspect - 1:
                    end = '\n'
                print(msg, end=end, flush=True)
            else:
                print(msg, end='\n')

    # ---------------------------
    # Reshape in case of same_spectrum
    if dinput['same_spectrum'] is True:
        nspect0 = dinput['same_spectrum_nspect']

        def reshape_custom(aa, nspect0=nspect0):
            return aa.reshape((nspect0, int(aa.size/nspect0)))

        nlamb = int(lamb.size / nspect0)
        nlines = int((sol_detail.shape[1]-1)/nspect0)
        lamb = lamb[:nlamb]

        nxbis = int(
            dind['bck_amp']['x'].size
            + dind['bck_rate']['x'].size
            + (dind['amp']['x'].size + dind['width']['x'].size)/nspect0
            + dind['shift']['x'].size
        )
        if dinput['double'] is not False:
            if dinput['double'] is True:
                nxbis += 2
            else:
                nxbis += (
                    dinput['double'].get('dratio') is not None
                    + dinput['double'].get('dshift') is not None
                )
        nba = dind['bck_amp']['x'].size
        nbr = dind['bck_rate']['x'].size
        nb = nba+nbr
        na = int(dind['amp']['x'].size/nspect0)
        nw = int(dind['width']['x'].size/nspect0)
        ns = dind['shift']['x'].size
        x2 = np.full((nspect0, nxbis), np.nan)
        x2[:, :nba] = sol_x[0, dind['bck_amp']['x']][None, :]
        x2[:, nba:nbr] = sol_x[0, dind['bck_rate']['x']][None, :]
        x2[:, nb:nb+na] = reshape_custom(sol_x[0, dind['amp']['x']])
        x2[:, nb+na:nb+na+nw] = reshape_custom(sol_x[0, dind['width']['x']])
        x2[:, nb+na+nw:nb+na+nw+ns] = sol_x[:, dind['shift']['x']]
        if dinput['double'] is True:
            x2[:, dind['dratio']['x']] = sol_x[:, dind['dratio']['x']]
            x2[:, dind['dshift']['x']] = sol_x[:, dind['dshift']['x']]
        import pdb; pdb.set_trace()     # DB
        sol_x = x2

    # ---------------------------
    # Isolate dratio and dshift
    dratio, dshift = None, None
    if dinput['double'] is not False:
        if dinput['double'] is True:
            dratio = (
                sol_x[:, dind['dratio']['x']] * scales[:, dind['dratio']['x']]
            )
            dshift = (
                sol_x[:, dind['dshift']['x']] * scales[:, dind['dshift']['x']]
            )
        else:
            if dinput['double'].get('dratio') is None:
                dratio = (
                    sol_x[:, dind['dratio']['x']]
                    * scales[:, dind['dratio']['x']]
                )
            else:
                dratio = np.full((nspect,), dinput['double']['dratio'])

            if dinput['double'].get('dshift') is None:
                dshift = (
                    sol_x[:, dind['dshift']['x']]
                    * scales[:, dind['dshift']['x']]
                )
            else:
                dshift = np.full((nspect,), dinput['double']['dshift'])

    if verbose > 0:
        dt = (dtm.datetime.now() - t0).total_seconds()
        msg = (
            "Total computation time:"
            + "\t{} s for {} spectra ({} s per spectrum)".format(
                round(dt, ndigits=3),
                nspect,
                round(dt/nspect, ndigits=3),
            )
        )
        print(msg)

    # ---------------------------
    # Format output as dict
    dfit = {
        'dinput': dinput,
        'scales': scales, 'x0': x0, 'bounds': bounds,
        'jac': jac, 'sol_x': sol_x,
        'dratio': dratio, 'dshift': dshift,
        'indx': indx,
        'time': time, 'success': success,
        'validity': validity,
        'dvalidity': _DVALIDITY,
        'errmsg': np.array(errmsg),
        'cost': cost, 'nfev': nfev, 'msg': np.array(message),
        'const': const,
        'xtol': xtol, 'ftol': ftol, 'gtol': gtol,
    }
    return dfit


def multigausfit2d_from_dlines(
    dinput=None,
    method=None, tr_solver=None, tr_options=None,
    xtol=None, ftol=None, gtol=None,
    max_nfev=None, chain=None, verbose=None,
    loss=None, jac=None,
    strict=None,
):
    """ Solve multi_gaussian fit in 1d from dlines

    If double is True, all lines are double with common shift and ratio

    Unknowns are:
        x = [bck, w0, v0, c00, c01, ..., c0n, w1, v1, c10, c11, ..., c1N, ...]

        - bck : constant background
        - wi  : spectral width of a group of lines (ion): wi^2 = 2kTi / m*c**2
                This way, it is dimensionless
        - vni : normalised velicity of the ion: vni = vi / c
        - cij : normalised coef (intensity) of line: cij = Aij

    Scaling is done so each quantity is close to unity:
        - bck: np.mean(data[data < mean(data)/2])
        - wi : Dlamb / 20
        - vni: 10 km/s
        - cij: np.mean(data)

    """

    # ---------------------------
    # Check format options
    (
        chain, method, tr_solver, tr_options,
        xtol, ftol, gtol, loss, max_nfev,
        verbose, verbscp, strict,
    ) = _checkformat_options(
         chain, method, tr_solver, tr_options,
         xtol, ftol, gtol, loss, max_nfev, verbose, strict,
    )

    # ---------------------------
    # Load dinput if necessary
    dinput = _checkformat_dinput(dinput)
    dprepare, dind = dinput['dprepare'], dinput['dind']
    nspect = dprepare['data'].shape[0]

    # ---------------------------
    # lamb and phi (symmetry axis?)
    lamb = dprepare['lamb']
    if dinput['symmetry'] is True:
        phi = np.abs(dprepare['phi'] - np.nanmean(dinput['symmetry_axis']))
    else:
        phi = dprepare['phi']

    # ---------------------------
    # Get scaling, x0, bounds from dict
    scales = _dict2vector_dscalesx0bounds(
        dd=dinput['dscales'], dd_name='dscales', dinput=dinput,
    )
    # muliply amplitudes by scales bs in [0, 1]
    scales[:, dinput['dind']['amp']['x']] *= dinput['dscales']['bs'][..., None]

    x0 = _dict2vector_dscalesx0bounds(
        dd=dinput['dx0'], dd_name='dx0', dinput=dinput,
    )
    boundmin = _dict2vector_dscalesx0bounds(
        dd=dinput['dbounds']['min'], dd_name="dbounds['min']", dinput=dinput,
    )
    boundmax = _dict2vector_dscalesx0bounds(
        dd=dinput['dbounds']['max'], dd_name="dbounds['max']", dinput=dinput,
    )
    bounds = np.array([boundmin[0, :], boundmax[0, :]])

    # ---------------------------
    # Prepare const

    const = _dict2vector_dscalesx0bounds(
        dd=dinput['dconstants'], dd_name='dconstants', dinput=dinput,
    )

    # -----------------------
    # Prepare indbs

    # bck_amp, bck_rate, all amp, width and shift are affected
    indbs = dinput['valid']['indbs']
    indbsfull = ~np.tile(indbs, dinput['dind']['nvar_bs'])
    ndiff = const.shape[1] - indbsfull.shape[1]
    assert ndiff <= 2
    if ndiff > 0:
        indbsfull = np.concatenate(
            (indbsfull, np.zeros((nspect, ndiff), dtype=bool)),
            axis=1,
        )
    const[indbsfull] = 1.

    # ---------------------------
    # Separate free from constant parameters

    # const = nan => x valid
    indx = np.isnan(const)
    x0[~indx] = const[~indx] / scales[~indx]

    # -----------------------
    # Prepare flattened data

    indok_all = np.any(dprepare['indok_bool'], axis=0)
    indok_flat = dprepare['indok_bool'][:, indok_all].reshape((nspect, -1))
    data_flat = dprepare['data'][:, indok_all].reshape((nspect, -1))
    phi_flat = phi[indok_all].ravel()
    lamb_flat = lamb[indok_all].ravel()

    # prepare lambrel and lambn
    lambrel_flat = lamb_flat - dinput['lambmin_bck']
    lambn_flat = lamb_flat[:, None] / dinput['lines'][None, ...]

    # jac0
    jac0 = np.zeros((phi_flat.size, dind['sizex']), dtype=float)

    # libs
    libs = np.array([
        (phi_flat >= dinput['knots_mult'][ii])
        & (phi_flat <= dinput['knots_mult'][ii + dinput['nknotsperbs'] - 1])
        for ii in range(dinput['nbs'])
    ])

    # ---------------------------
    # Get function, cost function and jacobian

    (
        func_cost, func_jac,
    ) = _funccostjac.multigausfit2d_from_dlines_funccostjac(
        phi_flat=phi_flat,
        dinput=dinput, dind=dind, jac=jac,
    )[2:]

    # ---------------------------
    # Prepare output
    sol_x = np.full((nspect, dind['sizex']), np.nan)
    sol_x[~indx] = const[~indx] / scales[~indx]
    success = np.full((nspect,), np.nan)
    time = np.full((nspect,), np.nan)
    cost = np.full((nspect,), np.nan)
    nfev = np.full((nspect,), np.nan)
    validity = np.zeros((nspect,), dtype=int)
    saturated = np.zeros((nspect, dind['sizex']), dtype=bool)
    message = ['' for ss in range(nspect)]
    errmsg = ['' for ss in range(nspect)]

    indamp = np.zeros((dind['sizex'],), dtype=bool)
    indamp[dinput['dind']['amp']['x'].T.ravel()] = True

    # Prepare msg
    if verbose in [1, 2]:
        col = np.char.array(['Spect', 'time (s)', 'cost',
                             'nfev', 'njev', 'msg'])
        maxl = max(np.max(np.char.str_len(col)), 10)
        msg = '\n'.join([
            ' '.join([cc.ljust(maxl) for cc in col]),
            ' '.join(['-'*maxl]*6),
        ])
        print(msg)

    # ------------
    # Main loop

    end = '\r'
    t0 = dtm.datetime.now()     # DB
    for ii in range(nspect):

        if verbose == 3:
            msg = "\nSpect {} / {}".format(ii+1, nspect)
            print(msg)

        try:
            dti = None
            t0i = dtm.datetime.now()     # DB
            if not dinput['valid']['indt'][ii]:
                validity[ii] = -1
                continue

            deltab = bounds[1, indx[ii, :]] - bounds[0, indx[ii, :]]

            # optimization
            res = scpopt.least_squares(
                func_cost,
                x0[ii, indx[ii, :]],
                jac=func_jac,
                bounds=bounds[:, indx[ii, :]],
                method=method,
                ftol=ftol,
                xtol=xtol,
                gtol=gtol,
                x_scale=1.0,
                f_scale=1.0,
                loss=loss,
                diff_step=None,
                tr_solver=tr_solver,
                tr_options=tr_options,
                jac_sparsity=None,
                max_nfev=max_nfev,
                verbose=verbscp,
                args=(),
                kwargs={
                    'indx': indx[ii, :],
                    'data_flat': data_flat[ii, indok_flat[ii]],
                    'scales': scales[ii, :],
                    'const': const[ii, ~indx[ii, :]],
                    'indok_flat': indok_flat[ii],
                    'phi_flat': phi_flat[indok_flat[ii]],
                    'lambrel_flat': lambrel_flat[indok_flat[ii]],
                    'lambn_flat': lambn_flat[indok_flat[ii], :],
                    'jac0': jac0[indok_flat[ii], :],
                    'libs': [ibs[indok_flat[ii]] for ibs in libs],
                }
            )
            dti = (dtm.datetime.now() - t0i).total_seconds()

            if chain is True and ii < nspect-1:
                x0[ii+1, indx[ii, :]] = res.x

            # cost, message, time
            success[ii] = res.success
            cost[ii] = res.cost
            nfev[ii] = res.nfev
            message[ii] = res.message
            time[ii] = round(
                (dtm.datetime.now()-t0i).total_seconds(),
                ndigits=3,
            )
            sol_x[ii, indx[ii, :]] = res.x

            # detect saturated values
            # amp at 0 are ok
            saturated[ii, indx[ii, :]] = (
                res.x > bounds[1, indx[ii, :]] - deltab*1e-4
            )
            saturated[ii, indx[ii, :] & indamp] |= (
                res.x[indamp[indx[ii, :]]] < 0.
            )
            saturated[ii, indx[ii, :] & (~indamp)] |= (
                res.x[(~indamp)[indx[ii, :]]]
                < (
                    bounds[0, indx[ii, :] & (~indamp)]
                    + 1.e-4 * deltab[(~indamp)[indx[ii, :]]]
                )
            )

        except Exception as err:
            if strict:
                raise err
            else:
                errmsg[ii] = str(err)
                validity[ii] = -2

        # verbose
        if verbose in [1, 2]:
            if validity[ii] == 0:
                col = np.char.array([
                    '{} / {}'.format(ii+1, nspect),
                    '{}'.format(dti),
                    '{:5.3e}'.format(res.cost),
                    str(res.nfev),
                    str(res.njev),
                    res.message,
                ])
            else:
                col = np.char.array([
                    '{} / {}'.format(ii+1, nspect),
                    '{}'.format(dti),
                    ' - ', ' - ', ' - ',
                    errmsg[ii],
                ])
            msg = ' '.join([cc.ljust(maxl) for cc in col])
            if verbose == 1:
                if ii == nspect - 1:
                    end = '\n'
                print(msg, end=end, flush=True)
            else:
                print(msg, end='\n')

    # ---------------
    # Display saturated values
    dsat = None
    if np.any(saturated):
        lksat = [
            'bck_amp', 'bck_rate',
            'amp', 'width', 'shift',
            'dshift', 'dratio',
        ]
        dsat = {
            k0: {
                'ind': np.sum(saturated[:, dind[k0]['x']], axis=1),
            }
            for k0 in lksat
            if dind.get(k0) is not None
            and np.any(saturated[:, dind[k0]['x']])
        }

        for k0 in dsat.keys():
            if k0 in ['amp', 'width', 'shift']:
                dsat[k0]['str'] = {}
                for ik, k1 in enumerate(dinput[k0]['keys']):
                    indk1 = dsat[k0]['ind'][:, ik] > 0
                    if np.any(indk1):
                        dsat[k0]['str'][k1] = (
                            indk1.sum(),
                            np.mean(dsat[k0]['ind'][indk1, ik]),
                        )
            else:
                indk1 = dsat[k0]['ind'][:, 0] > 0
                if np.any(indk1):
                    dsat[k0]['str'] = {
                        '': (indk1.sum(), np.mean(dsat[k0]['ind'][indk1, 0])),
                    }

        lstr = [
            "\n".join([
                f"\t{k0} {k1}: {v1[0]} / {nspect} "
                f"(mean {v1[1]} / {dinput['nbs']} bsplines)"
                for k1, v1 in v0['str'].items()
            ])
            for k0, v0 in dsat.items()
        ]
        msg = (
            "The following variables seem to have saturated:\n"
            + "\n".join(lstr)
        )
        print(msg)

    # ---------------------------
    # Isolate dratio and dshift
    dratio, dshift = None, None
    if dinput['double'] is not False:
        if dinput['double'] is True:
            dratio = (
                sol_x[:, dind['dratio']['x']] * scales[:, dind['dratio']['x']]
            )
            dshift = (
                sol_x[:, dind['dshift']['x']] * scales[:, dind['dshift']['x']]
            )
        else:
            if dinput['double'].get('dratio') is None:
                dratio = (
                    sol_x[:, dind['dratio']['x']]
                    * scales[:, dind['dratio']['x']]
                )
            else:
                dratio = np.full((nspect,), dinput['double']['dratio'])

            if dinput['double'].get('dshift') is None:
                dshift = (
                    sol_x[:, dind['dshift']['x']]
                    * scales[:, dind['dshift']['x']]
                )
            else:
                dshift = np.full((nspect,), dinput['double']['dshift'])

    if verbose > 0:
        dt = (dtm.datetime.now() - t0).total_seconds()
        msg = (
            "Total computation time:"
            + "\t{} s for {} spectra ({} s per spectrum)".format(
                round(dt, ndigits=3),
                nspect,
                round(dt/nspect, ndigits=3)
            )
        )
        print(msg)

    # ---------------------------
    # Format output as dict
    dfit = {
        'dinput': dinput,
        'scales': scales, 'x0': x0, 'bounds': bounds,
        'jac': jac, 'sol_x': sol_x,
        'dratio': dratio, 'dshift': dshift,
        'indx': indx,
        'time': time, 'success': success,
        'validity': validity,
        'dvalidity': _DVALIDITY,
        'errmsg': np.array(errmsg),
        'cost': cost, 'nfev': nfev, 'msg': np.array(message),
        'phi': phi,
        'const': const,
        'dsat': dsat,
        'xtol': xtol, 'ftol': ftol, 'gtol': gtol,
    }
    return dfit


###########################################################
###########################################################
#
#   Main fit functions
#
###########################################################
###########################################################


def fit1d(
    dinput=None,
    method=None, tr_solver=None, tr_options=None,
    xtol=None, ftol=None, gtol=None,
    max_nfev=None, loss=None, chain=None,
    jac=None, verbose=None, strict=None,
    # saving
    save=None,
    name=None,
    path=None,
    # extract for plotting
    amp=None, ratio=None,
    Ti=None, width=None,
    vi=None, shift=None,
    sol_total=None, sol_detail=None, sol_lamb=None,
    # plotting
    plot=None,
    showonly=None,
    fs=None, wintit=None, tit=None, dmargin=None,
    return_dax=None,
):

    # ----------------------
    # Check / format
    if showonly is None:
        showonly = False
    if save is None:
        save = False
    if plot is None:
        plot = False
    if return_dax is None:
        return_dax = False

    # ----------------------
    # Get dinput for 1d fitting from dlines, dconstraints, dprepare...
    if not isinstance(dinput, dict):
        msg = ("Please provide a properly formatted dict of inputs!\n"
               + "fit1d() needs the problem to be given as a dinput dict\n"
               + "  => Use dinput = fit1d_dinput()")
        raise Exception(msg)

    # ----------------------
    # Perform 2d fitting
    if showonly is True:
        msg = "TBF: lambfit and spect1d not defined"
        raise Exception(msg)

        dfit1d = {
            'shift': np.zeros((1, dinput['nlines'])),
            'coefs': np.zeros((1, dinput['nlines'])),
            'lamb': lambfit,
            'data': spect1d,
            'double': False,
            'Ti': False,
            'vi': False,
            'ratio': None,
        }
    else:
        dfit1d = multigausfit1d_from_dlines(
            dinput=dinput,
            method=method,
            max_nfev=max_nfev,
            tr_solver=tr_solver,
            tr_options=tr_options,
            xtol=xtol,
            ftol=ftol,
            gtol=gtol,
            loss=loss,
            chain=chain,
            verbose=verbose,
            jac=jac,
            strict=strict,
        )

    # ----------------------
    # Optional saving
    if save is True:
        if name is None:
            name = 'custom'
        name = 'TFS_fit1d_doutput_{}_{}_tol{}.npz'.format(
            name, dinput['method'], dinput['xtol'],
        )
        if not name.endswith('.npz'):
            name = name + '.npz'
        if path is None:
            path = os.getcwd()
        pfe = os.path.join(os.path.abspath(path), name)
        np.savez(pfe, **dfit1d)
        print(f"Saved in:\n\t{pfe}")

    # ----------------------
    # Optional plotting
    if plot is True:
        dout = fit1d_extract(
            dfit1d,
            amp=amp, ratio=ratio,
            Ti=Ti, width=width,
            vi=vi, shift=shift,
            sol_total=sol_total,
            sol_detail=sol_detail,
            sol_lamb=sol_lamb,
        )
        # TBF
        dax = _plot.plot_fit1d(
            dfit1d=dfit1d, dout=dout, showonly=showonly,
            fs=fs, dmargin=dmargin,
            tit=tit, wintit=wintit,
        )

    # ----------------------
    # return
    if return_dax is True:
        return dfit1d, dax
    else:
        return dfit1d


def fit2d(
    dinput=None,
    method=None, tr_solver=None, tr_options=None,
    xtol=None, ftol=None, gtol=None,
    max_nfev=None, loss=None, chain=None,
    jac=None,
    verbose=None, strict=None,
    # saving
    save=None,
    name=None,
    path=None,
    # extract for plotting
    amp=None, ratio=None,
    Ti=None, width=None,
    vi=None, shift=None,
    sol_total=None, sol_detail=None, sol_lamb_phi=None,
    # plotting
    plot=None,
    showonly=None,
    fs=None, wintit=None, tit=None, dmargin=None,
    return_dax=None,
):

    # ----------------------
    # Check / format
    if showonly is None:
        showonly = False
    if save is None:
        save = False
    if plot is None:
        plot = False
    if return_dax is None:
        return_dax = False

    # ----------------------
    # Get dinput for 1d fitting from dlines, dconstraints, dprepare...
    if not isinstance(dinput, dict):
        msg = ("Please provide a properly formatted dict of inputs!\n"
               + "fit2d() needs the problem to be given as a dinput dict\n"
               + "  => Use dinput = fit2d_dinput()")
        raise Exception(msg)

    # ----------------------
    # Perform 2d fitting
    if showonly is True:
        msg = "TBF: lambfit and spect1d not defined"
        raise Exception(msg)

        dfit2d = {
            'shift': np.zeros((1, dinput['nlines'])),
            'coefs': np.zeros((1, dinput['nlines'])),
            'lamb': None,
            'phi': None,
            'data': spect1d,
            'double': False,
            'Ti': False,
            'vi': False,
            'ratio': None,
        }
    else:
        dfit2d = multigausfit2d_from_dlines(
            dinput=dinput,
            method=method,
            tr_solver=tr_solver,
            tr_options=tr_options,
            xtol=xtol,
            ftol=ftol,
            gtol=gtol,
            max_nfev=max_nfev,
            chain=chain,
            verbose=verbose,
            loss=loss,
            jac=jac,
            strict=strict,
        )

    # ----------------------
    # Optional saving
    if save is True:
        if name is None:
            name = 'custom'
        name = 'TFS_fit2d_doutput_{}_nbs{}_{}_tol{}.npz'.format(
            name, dinput['nbs'], dinput['method'], dinput['xtol'],
        )
        if not name.endswith('.npz'):
            name = name + '.npz'
        if path is None:
            path = os.getcwd()
        pfe = os.path.join(os.path.abspath(path), name)
        np.savez(pfe, **dfit2d)
        print(f"Saved in:\n\t{pfe}")

    # ----------------------
    # Optional plotting
    if plot is True:
        dout = fit2d_extract(
            dfit2d,
            amp=amp, ratio=ratio,
            Ti=Ti, width=width,
            vi=vi, shift=shift,
            sol_total=sol_total,
            sol_detail=sol_detail,
            sol_lamb_phi=sol_lamb_phi,
        )
        # TBF
        dax = _plot.plot_fit2d(
            dfit2d=dfit2d, dout=dout, showonly=showonly,
            fs=fs, dmargin=dmargin,
            tit=tit, wintit=wintit,
        )

    # ----------------------
    # return
    if return_dax is True:
        return dfit2d, dax
    else:
        return dfit2d


###########################################################
###########################################################
#
#   Plot fitted data from pre-computed dict of fitted results
#
###########################################################
###########################################################

def fit2d_plot(dout=None):

    # ----------------------
    # Optional plotting
    if plot is True:
        if plotmode is None:
            plotmode = 'transform'
        if indspect is None:
            indspect = 0

        if spect1d is not None:
            # Compute lambfit / phifit and spectrum1d
            if nlambfit is None:
                nlambfit = 200
            ((spect1d, fit1d), lambfit,
             phifit, _, phiminmax) = self._calc_spect1d_from_data2d(
                 [dataflat[indspect, :], dfit2d['sol_tot'][indspect, :]],
                 lambflat, phiflat,
                 nlambfit=nlambfit, nphifit=10,
                 spect1d=spect1d, mask=None, vertsum1d=False)
        else:
            fit1d, lambfit, phiminmax = None, None, None

        dax = _plot_optics.CrystalBragg_plot_data_fit2d(
            xi=xi, xj=xj, data=dfit2d['data'],
            lamb=dfit2d['lamb'], phi=dfit2d['phi'], indspect=indspect,
            indok=indok, dfit2d=dfit2d,
            dax=dax, plotmode=plotmode, angunits=angunits,
            cmap=cmap, vmin=vmin, vmax=vmax,
            spect1d=spect1d, fit1d=fit1d,
            lambfit=lambfit, phiminmax=phiminmax,
            dmargin=dmargin, tit=tit, wintit=wintit, fs=fs)
    return dax


###########################################################
###########################################################
#
#           1d vertical fitting for noise analysis
#
###########################################################
###########################################################


def get_noise_costjac(deg=None, nbsplines=None, dbsplines=None, phi=None,
                      phiminmax=None, symmetryaxis=None, sparse=None):

    if sparse is None:
        sparse = False

    if dbsplines is None:
        dbsplines = multigausfit2d_from_dlines_dbsplines(
            knots=None, deg=deg, nbsplines=nbsplines,
            phimin=phiminmax[0], phimax=phiminmax[1],
            symmetryaxis=symmetryaxis)

    def cost(x,
             km=dbsplines['knots_mult'],
             deg=dbsplines['deg'],
             data=0., phi=phi):
        return scpinterp.BSpline(km, x, deg,
                                 extrapolate=False, axis=0)(phi) - data

    jac = np.zeros((phi.size, dbsplines['nbs']), dtype=float)
    km = dbsplines['knots_mult']
    kpb = dbsplines['nknotsperbs']
    lind = [(phi >= km[ii]) & (phi < km[ii+kpb-1])
            for ii in range(dbsplines['nbs'])]
    if sparse is True:
        def jac_func(x, jac=jac, km=km, data=None,
                     phi=phi, kpb=kpb, lind=lind):
            for ii in range(x.size):
                jac[lind[ii], ii] = scpinterp.BSpline.basis_element(
                    km[ii:ii+kpb], extrapolate=False)(phi[lind[ii]])
            return scpsparse.csr_matrix(jac)
    else:
        def jac_func(x, jac=jac, km=km, data=None,
                     phi=phi, kpb=kpb, lind=lind):
            for ii in range(x.size):
                jac[lind[ii], ii] = scpinterp.BSpline.basis_element(
                    km[ii:ii+kpb], extrapolate=False)(phi[lind[ii]])
            return jac
    return cost, jac_func


def _basic_loop(ilambu=None, ilamb=None, phi=None, data=None, mask=None,
                domain=None, nbs=None, dbsplines=None, nspect=None,
                method=None, tr_solver=None, tr_options=None, loss=None,
                xtol=None, ftol=None, gtol=None, max_nfev=None, verbose=None):

    # ---------------
    # Check inputs
    if method is None:
        method = _METHOD
    assert method in ['trf', 'dogbox', 'lm'], method
    if tr_solver is None:
        tr_solver = None
    if tr_options is None:
        tr_options = {}
    if xtol is None:
        xtol = _TOL2D['x']
    if ftol is None:
        ftol = _TOL2D['f']
    if gtol is None:
        gtol = _TOL2D['g']
    if loss is None:
        loss = _LOSS
    if max_nfev is None:
        max_nfev = None

    x0 = 1. - (2.*np.arange(nbs)/nbs - 1.)**2

    # ---------------
    # Prepare outputs
    dataint = np.full((nspect, ilambu.size), np.nan)
    fit = np.full(data.shape, np.nan)
    indsort = np.zeros((2, phi.size), dtype=int)
    indout_noeval = np.zeros(phi.shape, dtype=bool)
    chi2n = np.full((nspect, ilambu.size), np.nan)
    chi2_meandata = np.full((nspect, ilambu.size), np.nan)

    # ---------------
    # Main loop
    i0, indnan = 0, []
    for jj in range(ilambu.size):
        ind = ilamb == ilambu[jj]
        nind = ind.sum()
        isort = i0 + np.arange(0, nind)

        # skips cases with no points
        if not np.any(ind):
            continue

        inds = np.argsort(phi[ind])
        inds_rev = np.argsort(inds)
        indsort[0, isort] = ind.nonzero()[0][inds]
        indsort[1, isort] = ind.nonzero()[1][inds]

        phisort = phi[indsort[0, isort], indsort[1, isort]]
        datasort = data[:, indsort[0, isort], indsort[1, isort]]
        dataint[:, jj] = np.nanmean(datasort, axis=1)

        # skips cases with to few points
        indok = ~np.any(np.isnan(datasort), axis=0)
        if mask is not None:
            indok &= mask[indsort[0, isort], indsort[1, isort]]

        # Check there are enough phi vs bsplines
        indphimin = np.searchsorted(np.linspace(domain['phi']['minmax'][0],
                                                domain['phi']['minmax'][1],
                                                nbs + 1),
                                    phisort[indok])
        if np.unique(indphimin).size < nbs:
            indout_noeval[ind] = True
            continue
        indout_noeval[ind] = ~indok[inds_rev]

        # get bsplines func
        func_cost, func_jac = get_noise_costjac(phi=phisort[indok],
                                                dbsplines=dbsplines,
                                                sparse=False,
                                                symmetryaxis=False)
        for tt in range(nspect):
            if verbose > 0:
                msg = ("\tlambbin {} / {}".format(jj+1, ilambu.size)
                       + "    "
                       + "time step = {} / {}".format(tt+1, nspect))
                print(msg.ljust(50), end='\r', flush=True)

            if dataint[tt, jj] == 0.:
                continue

            datai = datasort[tt, indok] / dataint[tt, jj]
            res = scpopt.least_squares(
                func_cost, x0, jac=func_jac,
                method=method, ftol=ftol, xtol=xtol, gtol=gtol,
                x_scale='jac', f_scale=1.0, loss=loss, diff_step=None,
                tr_solver=tr_solver, tr_options={}, jac_sparsity=None,
                max_nfev=max_nfev, verbose=0, args=(),
                kwargs={'data': datai})

            # Store in original shape
            fit[tt, ind] = (
                func_cost(res.x, phi=phisort, data=0.)
                * dataint[tt, jj]
            )[inds_rev]
            chi2_meandata[tt, jj] = np.nanmean(fit[tt, ind])
            chi2n[tt, jj] = np.nanmean(func_cost(x=res.x, data=datai)**2)

        i0 += nind
        indnan.append(i0)
    return (fit, dataint, indsort, np.array(indnan), indout_noeval,
            chi2n, chi2_meandata)


def noise_analysis_2d(
    data, lamb, phi, mask=None, margin=None, valid_fraction=None,
    deg=None, knots=None, nbsplines=None, nxerrbin=None,
    nlamb=None, loss=None, max_nfev=None,
    xtol=None, ftol=None, gtol=None,
    method=None, tr_solver=None, tr_options=None,
    verbose=None, plot=None,
    ms=None, dcolor=None,
    dax=None, fs=None, dmargin=None,
    wintit=None, tit=None, sublab=None,
    save_fig=None, name_fig=None, path_fig=None, fmt=None,
    return_dax=None,
):

    # -------------
    # Check inputs
    if not isinstance(nbsplines, int):
        msg = "Please provide a (>0) integer value for nbsplines"
        raise Exception(msg)

    if deg is None:
        deg = 2
    if plot is None:
        plot = True
    if verbose is None:
        verbose = 1
    if return_dax is None:
        return_dax = False

    c0 = lamb.shape == phi.shape == data.shape[1:]
    if c0 is not True:
        msg = (
            "input data, lamb, phi are non-conform!\n"
            + "\t- expected lamb.shape == phi.shape == data.shape[1:]\n"
            + "\t- provided:\n"
            + "\t\tlamb.shape = {}\n".format(lamb.shape)
            + "\t\tphi.shape = {}\n".format(phi.shape)
            + "\t\tdata.shape = {}\n".format(data.shape)
        )
        raise Exception(msg)

    nspect = data.shape[0]
    domain = {'lamb': {'minmax': [np.nanmin(lamb), np.nanmax(lamb)]},
              'phi': {'minmax': [np.nanmin(phi), np.nanmax(phi)]}}

    if nlamb is None:
        if lamb.ndim == 2:
            nlamb = lamb.shape[0]
        else:
            msg = ("Please provide a value for nlamb (nb of bins)!")
            raise Exception(msg)
    nlamb = int(nlamb)

    # -------------
    # lamb binning
    lambedges = np.linspace(domain['lamb']['minmax'][0],
                            domain['lamb']['minmax'][1], nlamb+1)
    ilamb = np.searchsorted(lambedges, lamb)
    ilambu = np.unique(ilamb)

    # -------------
    # bspline dict and plotting utilities
    dbsplines = multigausfit2d_from_dlines_dbsplines(
        knots=None, deg=deg, nbsplines=nbsplines,
        phimin=domain['phi']['minmax'][0],
        phimax=domain['phi']['minmax'][1],
        symmetryaxis=False)

    # plotting utils
    bs_phi = np.linspace(domain['phi']['minmax'][0],
                         domain['phi']['minmax'][1], 101)
    bs_val = np.array([
        scpinterp.BSpline.basis_element(
            dbsplines['knots_mult'][ii:ii+dbsplines['nknotsperbs']],
            extrapolate=False)(bs_phi)
        for ii in range(nbsplines)]).T

    # -------------
    # Perform fits
    (fit, dataint, indsort, indnan, indout_noeval,
     chi2n, chi2_meandata) = _basic_loop(
        ilambu=ilambu, ilamb=ilamb, phi=phi, data=data, mask=mask,
        domain=domain, nbs=nbsplines, dbsplines=dbsplines, nspect=nspect,
        method=method, tr_solver=tr_solver, tr_options=tr_options, loss=loss,
        xtol=xtol, ftol=ftol, gtol=gtol,
        max_nfev=max_nfev, verbose=verbose)

    # -------------
    # Identify outliers with respect to noise model
    (mean, var, xdata, const,
     indout_var, _, margin, valid_fraction) = get_noise_analysis_var_mask(
         fit=fit, data=data, mask=(mask & (~indout_noeval)),
         margin=margin, valid_fraction=valid_fraction)

    # Safety check
    if mask is None:
        indout_mask = np.zeros(lamb.shape, dtype=bool)
    else:
        indout_mask = ~mask
    indout_noeval[~mask] = False
    indout_tot = np.array([~mask,
                           indout_noeval,
                           np.any(indout_var, axis=0)])
    c0 = np.all(np.sum(indout_tot.astype(int), axis=0) <= 1)
    if not c0:
        msg = "Overlapping indout!"
        raise Exception(msg)

    indin = ~np.any(indout_tot, axis=0)

    # -------------
    # output dict
    dnoise = {
        'data': data, 'phi': phi, 'fit': fit,
        'chi2n': chi2n, 'chi2_meandata': chi2_meandata, 'dataint': dataint,
        'domain': domain, 'indin': indin, 'indout_mask': indout_mask,
        'indout_noeval': indout_noeval, 'indout_var': indout_var,
        'mask': mask, 'ind_noeval': None,
        'indsort': indsort, 'indnan': np.array(indnan),
        'nbsplines': nbsplines, 'bs_phi': bs_phi, 'bs_val': bs_val,
        'deg': deg, 'lambedges': lambedges, 'deg': deg,
        'ilamb': ilamb, 'ilambu': ilambu,
        'var_mean': mean, 'var': var, 'var_xdata': xdata,
        'var_const': const, 'var_margin': margin,
        'var_fraction': valid_fraction}

    # Plot
    if plot is True:
        try:
            dax = _plot.plot_noise_analysis(
                dnoise=dnoise,
                ms=ms, dcolor=dcolor,
                dax=dax, fs=fs, dmargin=dmargin,
                wintit=wintit, tit=tit, sublab=sublab,
                save=save_fig, name=name_fig, path=path_fig, fmt=fmt)
        except Exception as err:
            msg = ("Plotting failed: {}".format(str(err)))
            warnings.warn(msg)
    if return_dax is True:
        return dnoise, dax
    else:
        return dnoise


def noise_analysis_2d_scannbs(
    data, lamb, phi, mask=None, nxerrbin=None,
    deg=None, knots=None, nbsplines=None, lnbsplines=None,
    nlamb=None, loss=None, max_nfev=None,
    xtol=None, ftol=None, gtol=None,
    method=None, tr_solver=None, tr_options=None,
    verbose=None, plot=None,
    dax=None, fs=None, dmargin=None,
    wintit=None, tit=None, ms=None, sublab=None,
    save_fig=None, name_fig=None, path_fig=None,
    fmt=None, return_dax=None,
):

    # -------------
    # Check inputs
    if lnbsplines is None:
        lnbsplines = np.arange(5, 21)
    else:
        lnbsplines = np.atleast_1d(lnbsplines).ravel().astype(int)
    if nbsplines is None:
        nbsplines = int(lnbsplines.size/2)
    if nbsplines is not None:
        nbsplines = np.unique(np.atleast_1d(nbsplines)).astype(int)
    nlnbs = lnbsplines.size
    if nxerrbin is None:
        nxerrbin = 100

    if deg is None:
        deg = 2
    if plot is None:
        plot = True
    if verbose is None:
        verbose = 1
    if return_dax is None:
        return_dax = False

    c0 = lamb.shape == phi.shape == data.shape[1:]
    if c0 is not True:
        msg = ("input data, lamb, phi are non-conform!\n"
               + "\t- expected lamb.shape == phi.shape == data.shape[1:]\n"
               + "\t- provided: ")
        raise Exception(msg)

    nspect = data.shape[0]
    domain = {'lamb': {'minmax': [np.nanmin(lamb), np.nanmax(lamb)]},
              'phi': {'minmax': [np.nanmin(phi), np.nanmax(phi)]}}

    if nlamb is None:
        if lamb.ndim == 2:
            nlamb = lamb.shape[0]
        else:
            msg = ("Please provide a value for nlamb (nb of bins)!")
            raise Exception(msg)
    nlamb = int(nlamb)

    # -------------
    # lamb binning
    lambedges = np.linspace(domain['lamb']['minmax'][0],
                            domain['lamb']['minmax'][1], nlamb+1)
    ilamb = np.searchsorted(lambedges, lamb)
    ilambu = np.unique(ilamb)

    # -------------
    # Perform fits
    xdata_edge = np.linspace(0, np.nanmax(data[:, mask]), nxerrbin+1)
    xdata = 0.5*(xdata_edge[1:] + xdata_edge[:-1])
    dataint = np.full((nspect, ilambu.size), np.nan)
    # fit = np.full(data.shape, np.nan)
    indsort = np.zeros((2, phi.size), dtype=int)
    # indout_noeval = np.zeros(phi.shape, dtype=bool)
    chi2n = np.full((nlnbs, nspect, ilambu.size), np.nan)
    chi2_meandata = np.full((nlnbs, nspect, ilambu.size), np.nan)
    const = np.full((nlnbs,), np.nan)
    mean = np.full((nlnbs, nxerrbin), np.nan)
    var = np.full((nlnbs, nxerrbin), np.nan)
    bs_phidata, bs_data, bs_fit, bs_indin = [], [], [], []
    for ii in range(lnbsplines.size):
        nbs = int(lnbsplines[ii])
        # -------------
        # bspline dict and plotting utilities
        dbsplines = multigausfit2d_from_dlines_dbsplines(
            knots=None, deg=deg, nbsplines=nbs,
            phimin=domain['phi']['minmax'][0],
            phimax=domain['phi']['minmax'][1],
            symmetryaxis=False,
        )

        # -------------
        # Perform fits
        if verbose > 0:
            msg = "nbs = {} ({} / {})".format(nbs, ii+1, lnbsplines.size)
            print(msg)
        (fiti, dataint, indsort, indnan, indout_noeval,
         chi2n[ii, ...], chi2_meandata[ii, ...]) = _basic_loop(
             ilambu=ilambu, ilamb=ilamb, phi=phi, data=data, mask=mask,
             domain=domain, nbs=nbs, dbsplines=dbsplines, nspect=nspect,
             method=method, tr_solver=tr_solver, tr_options=tr_options,
             loss=loss, xtol=xtol, ftol=ftol, gtol=gtol,
             max_nfev=max_nfev, verbose=verbose)

        if ii == 0:
            ind_intmax = np.unravel_index(np.argmax(dataint, axis=None),
                                          dataint.shape)

        if nbs in nbsplines:
            isi = np.split(indsort, indnan, axis=1)[ind_intmax[1]]
            bs_phidata.append(phi[isi[0], isi[1]])
            bs_data.append(data[ind_intmax[0], isi[0], isi[1]])
            bs_fit.append(fiti[ind_intmax[0], isi[0], isi[1]])
            indini = ~np.any(np.array([~mask, indout_noeval]), axis=0)
            bs_indin.append(indini[isi[0], isi[1]])

        # -------------
        # Identify outliers with respect to noise model
        (meani, vari, xdatai, consti,
         _, inderrui, _, _) = get_noise_analysis_var_mask(
             fit=fiti, data=data, xdata_edge=xdata_edge,
             mask=(mask & (~indout_noeval)),
             margin=None, valid_fraction=False)

        const[ii] = consti
        mean[ii, inderrui] = meani
        var[ii, inderrui] = vari

    # -------------
    # output dict
    dnoise_scan = {
        'data': data,
        'chi2n': chi2n, 'chi2_meandata': chi2_meandata, 'dataint': dataint,
        'domain': domain, 'lnbsplines': lnbsplines, 'nbsplines': nbsplines,
        'deg': deg, 'lambedges': lambedges, 'deg': deg,
        'ilamb': ilamb, 'ilambu': ilambu,
        'bs_phidata': bs_phidata, 'bs_data': bs_data,
        'bs_fit': bs_fit, 'bs_indin': bs_indin,
        'var_mean': mean, 'var': var, 'var_xdata': xdata,
        'var_const': const,
    }

    # Plot
    if plot is True:
        try:
            dax = _plot.plot_noise_analysis_scannbs(
                dnoise=dnoise_scan, ms=ms,
                dax=dax, fs=fs, dmargin=dmargin,
                wintit=wintit, tit=tit, sublab=sublab,
                save=save_fig, name=name_fig, path=path_fig, fmt=fmt)
        except Exception as err:
            msg = ("Plotting failed: {}".format(str(err)))
            warnings.warn(msg)

    if return_dax is True:
        return dnoise_scan, dax
    else:
        return dnoise_scan


def get_noise_analysis_var_mask(fit=None, data=None,
                                xdata_edge=None, nxerrbin=None,
                                valid_fraction=None,
                                mask=None, margin=None):
    if margin is None:
        margin = _SIGMA_MARGIN
    if valid_fraction is None:
        valid_fraction = False
    if nxerrbin is None:
        nxerrbin = 100

    err = fit - data
    if mask is None:
        mask = np.ones(err.shape[1:], dtype=bool)
    if xdata_edge is None:
        xdata_edge = np.linspace(0, np.nanmax(fit[:, mask]), nxerrbin)
    inderr = np.searchsorted(xdata_edge[1:-1], fit[:, mask])
    inderru = np.unique(inderr[~np.isnan(err[:, mask])])
    xdata = 0.5*(xdata_edge[1:] + xdata_edge[:-1])[inderru]
    mean = np.full((inderru.size,), np.nan)
    var = np.full((inderru.size,), np.nan)
    nn = np.full((inderru.size,), np.nan)
    for ii in range(inderru.size):
        ind = inderr == inderru[ii]
        indok = ~np.isnan(err[:, mask][ind])
        nn[ii] = np.sum(indok)
        mean[ii] = np.nanmean(err[:, mask][ind])
        var[ii] = nn[ii] * np.nanmean(err[:, mask][ind]**2) / (nn[ii] - 1)

    # fit sqrt on sigma (weight by log10 to take into account diff. nb. of pts)
    const = np.nansum(
        (np.log10(nn)*np.sqrt(var / xdata)) / np.nansum(np.log10(nn))
    )

    # indout
    indok = (~np.isnan(err)) & mask[None, ...]
    indout = np.zeros(err.shape, dtype=bool)
    indout[indok] = (np.abs(err[indok])
                     > margin*const*np.sqrt(np.abs(fit[indok])))
    if valid_fraction is not False:
        indout = np.sum(indout, axis=0)/float(indout.shape[0]) > valid_fraction
    return mean, var, xdata, const, indout, inderru, margin, valid_fraction





import inspect
import warnings


import numpy as np
import matplotlib as mpl


from . import _comp_optics


# ####################################################################
# ####################################################################
#               common routines for checking
# ####################################################################
# ####################################################################


def _check_dict_valid_keys(var=None, varname=None, valid_keys=None):
    c0 = (
        isinstance(var, dict)
        and all([
            isinstance(ss, str) and ss in valid_keys for ss in var.keys()
        ])
    )
    if not c0:
        msg = (
            """
            Arg {} must be:
                - dict
                - with valid keys in {}

            Provided:
                - type: {}
                - keys: {}
            """.format(
                varname,
                valid_keys,
                type(var),
                sorted(var.keys()) if isinstance(var, dict) else None)
        )
        raise Exception(msg)


def _check_flat1darray_size(var=None, varname=None, size=None, norm=None):

    # Default inputs
    if norm is None:
        norm = False

    # Return None of None
    if var is None:
        return None

    # Format to flat 1d array and check size
    var = np.atleast_1d(var).ravel()
    if var.size != size:
        msg = (
            """
            Var {} should be convertible to a 1d np.ndarray of size {}

            Provided:
                {}
            """.format(varname, size, var)
        )
        raise Exception(msg)

    # Normalize ?
    if norm is True:
        var = var / np.linalg.norm(var)
    return var


def _check_orthornormaldirect(e1=None, e2=None, var=None, varname=None):
    c0 = (
        np.abs(np.sum(e1*var)) < 1.e-12
        and np.abs(np.sum(e2*var)) < 1.e-12
        and np.linalg.norm(np.cross(e1, e2) - var) < 1.e-12
    )
    if not c0:
        msg = (
            """
            Unit vector basis (e1, e2, {0}) must be orthonormal direct!

            Provided:
                - norm(e1.{0}) = {}
                - norm(e2.{0}) = {}
                - norm((e1xe2) - {0}) = {}
            """.format(
                varname,
                np.abs(np.sum(e1*var)),
                np.abs(np.sum(e2*var)),
                np.linalg.norm(np.cross(e1, e2) - var),
            )
        )
        raise Exception(msg)


def _check_dict_unitvector(dd=None, dd_name=None):

    # Check all unit vectors
    for k0 in ['nout', 'nin', 'e1', 'e2']:
        dd[k0] = _check_flat1darray_size(
            var=dd.get(k0), varname=k0, size=3, norm=True)

    # Check consistency between unit vectors
    if dd['e1'] is not None:
        c0 = (
            dd['e2'] is not None
            and np.abs(np.sum(dd['e1']*dd['e2'])) < 1.e-12
        )
        if not c0:
            msg = (
                """
                If {0}['e1'] is provided, then:
                    - {0}['e2'] must be provided too
                    - {0}['e1'] must be perpendicular to {0}['e2']

                Provided:
                    {1}
                """.format(dd_name, dd['e2'])
            )
            raise Exception(msg)

        if dd['nout'] is not None:
            _check_orthornormaldirect(
                e1=dd['e1'], e2=dd['e2'],
                var=dd['nout'], varname='nout',
            )
        if dd['nin'] is not None:
            _check_orthornormaldirect(
                e1=dd['e1'], e2=dd['e2'],
                var=-dd['nin'], varname='-nin',
            )


# ####################################################################
# ####################################################################
#               check dgeom
# ####################################################################
# ####################################################################


def _checkformat_dgeom(dgeom=None, ddef=None, valid_keys=None):
    """
    Check the content of dgeom

    center, summit and rcurve

    extenthalf

    basis of unit vectors (e1, e2, nout) + nin
    """

    if dgeom is None:
        return

    # --------------------
    # Check dict integrity
    # --------------------

    # Check dict type and content (each key is a valid str)
    _check_dict_valid_keys(var=dgeom, varname='dgeom', valid_keys=valid_keys)

    # Set default values if any
    for kk in ddef.keys():
        dgeom[kk] = dgeom.get(kk, ddef[kk])

    # Set default values to None of not provided
    for kk in valid_keys:
        dgeom[kk] = dgeom.get(kk, None)

    # ------------------------------
    # Check each value independently
    # ------------------------------

    # Complementarity (center, rcurve) <=> (summit, rcurve)
    dgeom['center'] = _check_flat1darray_size(
        var=dgeom.get('center'), varname='center', size=3)
    dgeom['summit'] = _check_flat1darray_size(
        var=dgeom.get('summit'), varname='summit', size=3)

    lc = [dgeom['center'] is not None, dgeom['summit'] is not None]
    if not any(lc):
        msg = "Please provide at least dgeom['center'] and/or dgeom['summit']"
        raise Exception(msg)

    if dgeom['rcurve'] is None:
        msg = "Arg dgeom['rcurve'] must be convertible to a float"
        raise Exception(msg)
    dgeom['rcurve'] = float(dgeom['rcurve'])

    # Check dimensions (half extent)
    dgeom['extenthalf'] = _check_flat1darray_size(
        var=dgeom.get('extenthalf'), varname='extenthalf', size=2)

    # ----------------------------------
    # Add missing vectors and parameters
    # ----------------------------------

    if dgeom['e1'] is not None:
        if dgeom['nout'] is None:
            dgeom['nout'] = np.cross(dgeom['e1'], dgeom['e2'])
        if dgeom['nin'] is None:
            dgeom['nin'] = -dgeom['nout']
        if dgeom['center'] is None:
            dgeom['center'] = (dgeom['summit']
                               + dgeom['nin']*dgeom['rcurve'])
        if dgeom['summit'] is None:
            dgeom['summit'] = (dgeom['center']
                               + dgeom['nout']*dgeom['rcurve'])
    elif dgeom['center'] is not None and dgeom['summit'] is not None:
        if dgeom['nout'] is None:
            nout = (dgeom['summit'] - dgeom['center'])
            dgeom['nout'] = nout / np.linalg.norm(nout)
        if dgeom['e1'] is None:
            dgeom['e1'] = np.cross(np.r_[0, 0, 1], dgeom['nout'])
            dgeom['e1'] = dgeom['e1'] / np.linalg.norm(dgeom['e1'])
            msg = (
                "dgeom['e1'] was not provided!\n"
                + "  => setting e1 to horizontal by default!\n"
            )
            warnings.warn(msg)
        if dgeom['e2'] is None:
            dgeom['e2'] = np.cross(dgeom['nout'], dgeom['e1'])
            dgeom['e2'] = dgeom['e2'] / np.linalg.norm(dgeom['e2'])

    if dgeom['extenthalf'] is not None:
        if dgeom['Type'] == 'sph' and dgeom['Typeoutline'] == 'rect':
            ind = np.argmax(dgeom['extenthalf'])
            dphi = dgeom['extenthalf'][ind]
            sindtheta = np.sin(dgeom['extenthalf'][ind-1])
            dgeom['surface'] = 4.*dgeom['rcurve']**2*dphi*sindtheta

    # Check orthonormal direct basis
    _check_dict_unitvector(dd=dgeom, dd_name='dgeom')

    return dgeom


# ####################################################################
# ####################################################################
#               check dmat
# ####################################################################
# ####################################################################


def _checkformat_dmat(dmat=None, dgeom=None, ddef=None, valid_keys=None):
    """
    Check the content of dmat

    Crystal parameters : d, formula, density, lengths, angles, cut

    New basis of unit vectors due to miscut (e1, e2, nout)
    + nin + alpha, beta
    """

    if dmat is None:
        return

    # ---------------------
    # check dict integrity
    # ---------------------

    # Check dict typeand content (each key is a valid string)
    _check_dict_valid_keys(var=dmat, varname='dmat', valid_keys=valid_keys)

    # Set default values if any
    for kk in ddef.keys():
        dmat[kk] = dmat.get(kk, ddef[kk])

    # -------------------------------
    # check each value independently
    # -------------------------------

    # Check dimension of array and its size
    dmat['lengths'] = _check_flat1darray_size(
        var=dmat.get('lengths'), varname='lengths', size=3)
    dmat['angles'] = _check_flat1darray_size(
        var=dmat.get('angles'), varname='angles', size=3)

    if dmat['d'] is None:
        msg = "Arg dmat['d'] must be convertible to a float."
        raise Exception(msg)
    dmat['d'] = float(dmat['d'])

    if dmat['density'] is None:
        msg = "Arg dmat['density'] must be convertible to a float."
        raise Exception(msg)
    dmat['density'] = float(dmat['density'])

    if isinstance(dmat['formula'], str) is False:
        msg = (
            """
            Var {} must be a valid string.

            Provided
                - type: {}
            """.format('formula', type(dmat['formula']))
        )
        raise Exception(msg)

    if dmat.get('cut') is not None:
        dmat['cut'] = np.atleast_1d(dmat['cut']).ravel().astype(int)
        if dmat['cut'].size == 3:
            pass
        elif dmat['cut'].size == 4:
            pass
        else:
            msg = (
                "Var 'cut' should be convertible to a 1d array of size 3 or 4"
                f"\nProvided: {dmat.get('cut')}"
            )
            raise Exception(msg)

    # Check orthonormal direct basis
    _check_dict_unitvector(dd=dmat, dd_name='dmat')

    # Check all additionnal angles to define the new basis
    for k0 in ['alpha', 'beta']:
        dmat[k0] = _check_flat1darray_size(
            var=dmat.get(k0), varname=k0, size=1, norm=False)

    # -------------------------------------------------------------
    # Add missing vectors and parameters according to the new basis
    # -------------------------------------------------------------

    if all([dgeom[kk] is not None for kk in ['nout', 'e1', 'e2']]):

        # dict of value, comment, default value and type of
        #  alpha and beta angles in dmat
        dpar = {
            'alpha': {
                # 'alpha': alpha,
                'com': 'miscut amplitude',
                'default': 0.,
                'type': float,
            },
            'beta': {
                # 'beta': beta,
                'com': 'miscut orientation',
                'default': 0.,
                'type': float,
            },
        }

        # setting to default value if any is None
        lparNone = [aa for aa in dpar.keys() if dmat.get(aa) is None]

        # if any is None, assigning default value and send a warning message
        if len(lparNone) > 0:
            msg = "The following parameters were set to their default values:"
            for aa in lparNone:
                dmat[aa] = dpar[aa]['default']
                msg += "\n\t - {} = {} ({})".format(
                    aa, dpar[aa]['default'], dpar[aa]['com'],
                )
            warnings.warn(msg)

        # check conformity of type of angles
        lparWrong = []
        for aa in dpar.keys():
            try:
                dmat[aa] = float(dmat[aa])
            except Exception as err:
                lparWrong.append(aa)

        if len(lparWrong) > 0:
            msg = "The following parameters must be convertible to:"
            for aa in lparWrong:
                msg += "\n\t - {} = {} ({})".format(
                    aa, dpar[aa]['type'], type(dmat[aa]),
                )
            raise Exception(msg)

        # Check value of alpha
        if dmat['alpha'] < 0 or dmat['alpha'] > np.pi/2:
            msg = (
                "Arg dmat['alpha'] must be an angle (radians) in [0; pi/2]"
                + "\nProvided:\n\t{}".format(dmat['alpha'])
            )
            raise Exception(msg)

        if dmat['beta'] < -np.pi or dmat['beta'] > np.pi:
            msg = (
                "Arg dmat['beta'] must be an angle (radians) in [-pi; pi]"
                + "\nProvided:\n\t{}".format(dmat['beta'])
            )
            raise Exception(msg)

        # dict of value, comment, default value and type of unit vectors in
        #  dmat, computting default value of unit vectors, corresponding to
        #  dgeom values if angles are 0.
        dvec = {
            'e1': {
                # 'e1': e1,
                'com': 'unit vector (miscut)',
                'default': (
                    np.cos(dmat['alpha'])*(
                        np.cos(dmat['beta'])*dgeom['e1']
                        + np.sin(dmat['beta'])*dgeom['e2']
                    )
                    - np.sin(dmat['alpha'])*dgeom['nout']
                ),
                'type': float,
            },
            'e2': {
                # 'e2': e2,
                'com': 'unit vector (miscut)',
                'default': (
                    np.cos(dmat['beta'])*dgeom['e2']
                    - np.sin(dmat['beta'])*dgeom['e1']
                ),
                'type': float,
            },
            'nout': {
                # 'nout': nout,
                'com': 'outward unit vector (normal to non-parallel mesh)',
                'default': (
                    np.cos(dmat['alpha'])*dgeom['nout']
                    + np.sin(dmat['alpha']) * (
                        np.cos(dmat['beta'])*dgeom['e1']
                        + np.sin(dmat['beta'])*dgeom['e2']
                    )
                ),
                'type': float,
            }
        }

        # setting to default value if any is None
        lvecNone = [bb for bb in dvec.keys() if dmat.get(bb) is None]

        # if any is None, assigning default value
        if len(lvecNone) > 0:
            for bb in lvecNone:
                dmat[bb] = dvec[bb]['default']

        # check conformity of type of unit vectors
        lvecWrong = []
        for bb in dvec.keys():
            try:
                dmat[bb] = np.atleast_1d(dmat[bb]).ravel().astype(float)
                dmat[bb] = dmat[bb] / np.linalg.norm(dmat[bb])
                assert np.allclose(dmat[bb], dvec[bb]['default'])
            except Exception as err:
                lvecWrong.append(bb)

        if len(lvecWrong) > 0:
            msg = "The following parameters must be convertible to:"
            for bb in lvecWrong:
                msg += "\n\t - {} = {} ({})".format(
                    bb, dvec[bb]['type'], type(dmat[bb]),
                )
            msg += "\nAnd (nout , e1, e2) must be an orthonormal direct basis"
            raise Exception(msg)

        # Computation of unit vector nin
        dmat['nin'] = -dmat['nout']

    return dmat


# ####################################################################
# ####################################################################
#               check dbragg
# ####################################################################
# ####################################################################


def _checkformat_dbragg(dbragg=None, ddef=None, valid_keys=None, dmat=None):

    """
    rocking curve, wavelenght of reference and its bragg angle associated
    """
    if dbragg is None:
        dbragg = dict.fromkeys(valid_keys)

    # ---------------------
    # check dict integrity
    # ---------------------

    # Check type dict and content (each key is a valid string)
    _check_dict_valid_keys(var=dbragg, varname='dbragg', valid_keys=valid_keys)

    # Set default values when necessary
    for kk in ddef.keys():
        dbragg[kk] = dbragg.get(kk, ddef[kk])

    # ------------------------------------------------
    # Check braggref and lambref type and computation
    # ------------------------------------------------

    # Check braggref
    ltypes = [int, float, np.int_, np.float64]
    c0 = bool(
        type(dbragg.get('braggref')) in ltypes
        and dbragg['braggref'] >= 0
        and dbragg['braggref'] <= np.pi/2.
    )
    if not c0:
        msg = (
            """
            Var {} is not valid!
            Value should be in [0; pi/2]

            Provided: {}
            """.format('braggref', dbragg['braggref'])
            )
        raise Exception(msg)

    # Set default lambda if necessary
    user_prov = True
    if dbragg.get('lambref') is None:
        # set to bragg = braggref
        dbragg['lambref'] = _comp_optics.get_lamb_from_bragg(
            np.r_[dbragg['braggref']],
            dmat['d'],
            n=1,
        )[0]
        user_prov = False

    # Set default bragg angle if necessary
    braggref = _comp_optics.get_bragg_from_lamb(
                            np.r_[dbragg['lambref']], dmat['d'], n=1)[0]
    if np.isnan(braggref):
        lambok = []
        msg = (
            """
            Var {} is not valid!
            Please check your arguments to calculate the Bragg's law correctly!
            Provided:
                - crystal inter-plane d [m] = {}
                - wavelenght interval [m] : {}
                - lambref = {}
            """.format('lambref', dmat['d'], lambok, dbragg['lambref'])
        )
        raise Exception(msg)

    # Update braggref according to lambref, if lambref was user-provided
    if user_prov is True:
        dbragg['braggref'] = braggref

    # ------------------------------------------------
    # Check rocking curve value
    # ------------------------------------------------

    # Check type dict and content (each key is a valid string)
    drock = dbragg.get('rockingcurve')
    if drock is not None:
        lkeyok = [
            'sigma', 'deltad', 'Rmax', 'dangle', 'lamb',
            'value', 'type', 'source',
        ]
        _check_dict_valid_keys(var=drock, varname='drock', valid_keys=lkeyok)

        # check type, size and content of each key in drock
        # Rocking curve can be provided as:
        # - analytical form (Lorentzian-log)
        # - tabulated in 2d
        # - tabulated in 1d
        try:
            if drock.get('sigma') is not None:
                dbragg['rockingcurve']['sigma'] = float(drock['sigma'])
                dbragg['rockingcurve']['deltad'] = float(
                    drock.get('deltad', 0.),
                )
                dbragg['rockingcurve']['Rmax'] = float(drock.get('Rmax', 1.))
                dbragg['rockingcurve']['type'] = 'lorentz-log'

            elif drock.get('dangle') is not None:
                c2d = (drock.get('lamb') is not None
                       and drock.get('value').ndim == 2)
                if c2d:
                    if drock['value'].shape != (drock['dangle'].size,
                                                drock['lamb'].size):
                        msg = (
                            """ Tabulated 2d rocking curve should be:
                                shape = (dangle.size, lamb.size)
                            """)
                        raise Exception(msg)
                    dbragg['rockingcurve']['dangle'] = np.r_[drock['dangle']]
                    dbragg['rockingcurve']['lamb'] = np.r_[drock['lamb']]
                    dbragg['rockingcurve']['value'] = drock['value']
                    dbragg['rockingcurve']['type'] = 'tabulated-2d'

                else:
                    if drock.get('lamb') is None:
                        msg = (
                            """Please also specify the lamb for which
                                the rocking curve was tabulated""")
                        raise Exception(msg)
                    dbragg['rockingcurve']['lamb'] = float(drock['lamb'])
                    dbragg['rockingcurve']['dangle'] = np.r_[drock['dangle']]
                    dbragg['rockingcurve']['value'] = np.r_[drock['value']]
                    dbragg['rockingcurve']['type'] = 'tabulated-1d'

                if drock.get('source') is None:
                    msg = "Unknown source for the tabulated rocking curve!"
                    warnings.warn(msg)
                dbragg['rockingcurve']['source'] = drock.get('source')

        except Exception as err:
            msg = (
                """
                Provide the rocking curve as a dictionnary with either:
                    - parameters of a lorentzian in log10:
                      'sigma': float, 'deltad':float, 'Rmax': float
                    - tabulated (dangle, value) with source (url...):
                      'dangle': np.darray, 'value': np.darray, 'source': str
                """
            )
            raise Exception(msg)
    else:
        dbragg['rockingcurve'] = None
    return dbragg


# ####################################################################
# ####################################################################
#               check colors
# ####################################################################
# ####################################################################


def _checkformat_inputs_dmisc(color=None, ddef=None):
    if color is None:
        color = mpl.colors.to_rgba(ddef['dmisc']['color'])
    assert mpl.colors.is_color_like(color)
    return tuple(mpl.colors.to_rgba(color))


# ####################################################################
# ####################################################################
#               check synthetic diags
# ####################################################################
# ####################################################################


def _check_config_get_Ves(config=None, struct=None):

    if config.__class__.__name__ != 'Config':
        msg = (
            "Arg config must be a Config object "
            f"Provided:\n\t- config: {type(config)}"
        )
        raise Exception(msg)

    lok = list(config.dStruct['dObj']['Ves'].keys())
    if struct is None and len(lok) == 1:
        struct = lok[0]
    elif struct not in lok:
        msg = (
            "Arg struct must be the name of a StructIn in config!\n"
            f"Provided:\n\t- Available: {lok}\n\t- struct: {struct}"
        )
        raise Exception(msg)
    return struct


def _check_calc_signal_from_emissivity(
    emis=None,
    config=None,
    struct=None,
    lamb=None,
    det=None,
    binning=None,
):

    # config
    struct = _check_config_get_Ves(config=config, struct=struct)

    # lamb
    lamb = np.atleast_1d(lamb).ravel()

    # det
    assert det is not None

    # emis
    if not callable(emis):
        msg = "Arg emis must be a callable!"
        raise Exception(msg)

    argspec = inspect.getfullargspec(emis)
    c0 = (
        all([ss in argspec.args for ss in ['r', 'z', 'phi', 'lamb', 't']])
    )
    if not c0:
        msg = (
            "Arg emis must take at least the following keyword args:\n"
            "\t- r: major radius\n"
            "\t- z: height\n"
            "\t- phi: toroidal angle\n"
            "\t- lambda: wavelength\n"
            "\t- t: time\n"
            f"\nProvided: {argspec.args}"
        )
        raise Exception(msg)

    # binning
    if binning is None:
        binning = False
    if binning is not False:
        c0 = (
            isinstance(binning, dict)
            and sorted(binning.keys()) == ['xi', 'xj']
            and all([isinstance(v0, np.ndarray) for k0, v0 in binning.items()])
        )
        if c0:
            binning = (binning['xi'], binning['xj'])
        c0 = (
            hasattr(binning, '__iter__')
            and len(binning) == 2
            and all([isinstance(bb, np.ndarray) for bb in binning])
            and all([np.allclose(bb, np.unique(bb)) for bb in binning])
        )
        if not c0:
            msg = (
                "Arg binning must be either:\n"
                "\t- dict of keys ('xi', 'xj') with bin sorted edges arrays\n"
                "\t- list of 2 'xi' and 'xj' bin sorted edges arrays\n"
            )
            raise Exception(msg)

    return struct, lamb, binning


# -*- coding: utf-8 -*-
"""
Created on Thu Feb 27 17:27:55 2025

@author: dvezinet
"""


import numpy as np
import datastock as ds


from ..geom import Config, CamLOS1D


# ###########################################################
# ###########################################################
#                   Main
# ###########################################################


def main(
    coll=None,
    key=None,
    config=None,
    segment=None,
    allowed=None,
    excluded=None,
):

    # -----------------
    # check inputs
    # -----------------

    key, segment, lstruct = _check(
        coll=coll,
        key=key,
        config=config,
        segment=segment,
        allowed=allowed,
        excluded=excluded,
    )

    # -----------------
    # prepare
    # -----------------

    ptsx, ptsy, ptsz = coll.get_rays_pts(key)
    vx, vy, vz = coll.get_rays_vect(key)

    ptsx = ptsx[segment, ...]
    ptsy = ptsy[segment, ...]
    ptsz = ptsz[segment, ...]
    vx = vx[segment, ...]
    vy = vy[segment, ...]
    vz = vz[segment, ...]

    mask = np.isfinite(ptsx) & np.isfinite(vx)
    maskn = mask.nonzero()

    # -----------------
    # ray-tracing
    # -----------------

    # call legacy code  -- SOMETHING WRONG HERE
    cam = CamLOS1D(
        dgeom=(
            np.array([ptsx[mask], ptsy[mask], ptsz[mask]]),
            np.array([vx[mask], vy[mask], vz[mask]]),
        ),
        config=config,
        Name='',
        Diag='',
        Exp='',
        strict=False,
    )

    # -----------------
    # ectract
    # -----------------

    dtouch = cam.get_touch_dict()

    # intitialize
    itouch = -np.ones(ptsx.shape, dtype=int)

    for ii, ks in enumerate(lstruct):

        if ks not in dtouch.keys():
            continue

        sli = tuple([ind[dtouch[ks]['indok']] for ind in maskn])

        itouch[sli] = ii

    # -----------------
    # output
    # -----------------

    dout = {
        'lstruct': lstruct,
        'ind': itouch,
    }

    return dout


# ###########################################################
# ###########################################################
#                   check inputs
# ###########################################################


def _check(
    coll=None,
    key=None,
    config=None,
    segment=None,
    allowed=None,
    excluded=None,
):

    # -----------------
    # key
    # -----------------

    wrays = coll._which_rays
    lok = list(coll.dobj.get(wrays, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -----------------
    # segment
    # -----------------

    vx, vy, vz = coll.get_rays_vect(key)
    segment = int(ds._generic_check._check_var(
        segment, 'segment',
        types=(int, float),
        default=0,
        allowed=np.r_[range(vx.shape[0]), -1],
    ))

    # -----------------
    # config
    # -----------------

    if not isinstance(config, Config):
        msg = "Arg config must be a tofu Config instance!"
        raise Exception(msg)

    lok = [
        f"{ss.__class__.__name__}_{ss.Id.Name}" for ss in config.lStruct
    ]

    # -------------------
    # allowed vs excluded
    # -------------------

    lc = [allowed is not None, excluded is not None]
    if np.sum(lc) > 1:
        msg = "Please provide either excluded xor allowed, not both!"
        raise Exception(msg)

    # -----------------
    # lstruct
    # -----------------

    if allowed is None and excluded is None:
        lstruct = [
            f"{ss.__class__.__name__}_{ss.Id.Name}" for ss in config.lStruct
            if ss.get_visible()
        ]

    elif allowed is not None:

        if isinstance(allowed, str):
            allowed = [allowed]
        allowed = ds._generic_check._check_var_iter(
            allowed, 'allowed',
            allowed=lok,
            types=(list, tuple),
            types_iter=str,
        )

        lstruct = allowed

    else:
        if isinstance(excluded, str):
            excluded = [excluded]
        excluded = ds._generic_check._check_var_iter(
            excluded, 'excluded',
            allowed=lok,
            types=(list, tuple),
            types_iter=str,
        )

        lstruct = [ss for ss in lok if ss not in excluded]

    return key, segment, lstruct



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class02_Rays import Rays as Previous
from . import _class3_check as _check


__all__ = ['Aperture']


# #########################################################################
# #########################################################################
#                           Diagnostic
# #########################################################################


class Aperture(Previous):

    # _ddef = copy.deepcopy(ds.DataStock._ddef)
    # _ddef['params']['ddata'].update({
    #       'bsplines': (str, ''),
    # })
    # _ddef['params']['dobj'] = None
    # _ddef['params']['dref'] = None

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'aperture': [
            'dgeom.type',
            'dgeom.curve_r',
            'dgeom.area',
            'dgeom.outline',
            'dgeom.poly',
            'dgeom.cent',
            # 'dmisc.color',
        ],
    })

    def add_aperture(
        self,
        key=None,
        # 2d outline
        outline_x0=None,
        outline_x1=None,
        cent=None,
        extenthalf=None,
        # 3d outline
        poly_x=None,
        poly_y=None,
        poly_z=None,
        # normal vector
        nin=None,
        e0=None,
        e1=None,
        # curvature
        curve_r=None,
        make_planar=None,
        # dmisc
        color=None,
    ):
        """ Add an aperture

        Can be defined from:
            - 2d outline + 3d center + unit vectors (nin, e0, e1)
            - 3d polygon + nin

        Unit vectors will be checked and normalized
        If planar, area will be computed
        Outline will be made counter-clockwise

        """

        # check / format input
        dref, ddata, dobj = _check._add_surface3d(
            coll=self,
            key=key,
            which='aperture',
            which_short='ap',
            # 2d outline
            outline_x0=outline_x0,
            outline_x1=outline_x1,
            cent=cent,
            extenthalf=extenthalf,
            # 3d outline
            poly_x=poly_x,
            poly_y=poly_y,
            poly_z=poly_z,
            # normal vector
            nin=nin,
            e0=e0,
            e1=e1,
            # curvature
            curve_r=curve_r,
            make_planar=make_planar,
        )

        # dmisc
        key = list(dobj['aperture'].keys())[0]
        dobj['aperture'][key]['dmisc'] = _check._dmisc(
            key=key,
            color=color,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

    # ---------------
    # utilities
    # ---------------

    def get_as_dict(self, which=None, key=None):
        """ Return the desired object as a dict (input to some routines) """

        return _check._return_as_dict(
            coll=self,
            which=which,
            key=key,
        )



# -*- coding: utf-8 -*-


# Built-in
# import copy


# tofu
from ._class01_Plasma2D import Plasma2D as Previous
from . import _class2_check as _check
from . import _class02_touch as _touch
from . import _class02_sample as _sample
from . import _class02_tangency_radius as _tangency_radius
from . import _class02_single_point_camera as _single_point_cam
from . import _class2_plot as _plot
from . import _class2_sinogram as _sinogram
from . import _class02_save2stp as _save2stp
from . import _class8_plot


__all__ = ['Rays']


# ##############################################################
# ##############################################################
#                       Rays
# ##############################################################


class Rays(Previous):

    _which_rays = 'rays'
    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        _which_rays: [
            'shape',
            'ref',
            'pts',
            'lamb',
            'Rmin',
            'alpha',
            'dalpha',
            'dbeta',
        ],
    })

    def add_rays(
        self,
        key=None,
        # start points
        start_x=None,
        start_y=None,
        start_z=None,
        # wavelength
        lamb=None,
        # ref
        ref=None,
        # from pts
        pts_x=None,
        pts_y=None,
        pts_z=None,
        # from ray-tracing (vect + length or config or diag)
        vect_x=None,
        vect_y=None,
        vect_z=None,
        length=None,
        config=None,
        strict=None,
        # angles
        alpha=None,
        dalpha=None,
        dbeta=None,
        # reflections
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        diag=None,
        key_cam=None,
        # others
        **kwdargs,
    ):
        """ Add a set of rays

        Can be defined from:
            - pts: explicit
            - starting pts + vectors

        """

        # check / format input
        dref, ddata, dobj = _check._rays(
            coll=self,
            key=key,
            # start points
            start_x=start_x,
            start_y=start_y,
            start_z=start_z,
            # wavelength
            lamb=lamb,
            # ref
            ref=ref,
            # from pts
            pts_x=pts_x,
            pts_y=pts_y,
            pts_z=pts_z,
            # from ray-tracing (vect + length or config or diag)
            vect_x=vect_x,
            vect_y=vect_y,
            vect_z=vect_z,
            length=length,
            config=config,
            strict=strict,
            # angles
            alpha=alpha,
            dalpha=dalpha,
            dbeta=dbeta,
            # reflections
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            key_nseg=key_nseg,
            diag=diag,
            key_cam=key_cam,
            # kwdargs
            kwdargs=kwdargs,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

    # -----------------
    # remove
    # -----------------

    def remove_rays(self, key=None):
        return _check._remove(
            coll=self,
            key=key,
        )

    # --------------
    #  utilities
    # --------------

    def get_rays_start(
        self,
        key=None,
        key_cam=None,
    ):
        return _check._get_start(coll=self, key=key, key_cam=key_cam)

    def get_rays_pts(
        self,
        key=None,
        key_cam=None,
        segment=None,
    ):
        """ Return (ptsx, ptsy, ptsz) coordinates for rays

        segment = None => all segments returned, (nseg, ...) array
        segment = int => (2, ...) array returned
        """
        return _check._get_pts(
            coll=self,
            key=key,
            key_cam=key_cam,
            segment=segment,
        )

    def get_rays_vect(
        self,
        key=None,
        key_cam=None,
        norm=None,
        segment=None,
    ):
        """ Return (vx, vy, vz) coordinates for desired rays

        Optionally (default) normalized

        segment = None => all segments returned, (nseg-1, ...) array
        segment = int => (...,) array returned
        """
        return _check._get_vect(
            coll=self,
            key=key,
            key_cam=key_cam,
            norm=norm,
            segment=segment,
        )

    # --------------
    # touch
    # --------------

    def get_rays_touch_dict(
        self,
        key=None,
        config=None,
        segment=None,
        allowed=None,
        excluded=None,
    ):
        """ Compute which ray touches which solid in a config


        Parameters
        ----------
        key:    str
            key to an existing ray
        config: Config
            an existing Config instance

        Returns
        -------
        dout: dict
            output dict

        """
        return _touch.main(
            coll=self,
            key=key,
            config=config,
            segment=segment,
            allowed=allowed,
            excluded=excluded,
        )

    # --------------
    # discretizing
    # --------------

    def sample_rays(
        self,
        key=None,
        key_cam=None,
        res=None,
        mode=None,
        segment=None,
        ind_ch=None,
        radius_max=None,
        concatenate=None,
        return_coords=None,
    ):
        """ Return the sampled rays

        Parameters
        ----------
        key:        str
            key of the rays / diag los to sample
        res:        float
            sampling resolution
        mode:       str
            sampling mode
                - 'rel': relative, res is in [0, 1], 0.1 = 10 samples / segment
                - 'abs': absolute, res is a distance in m
        ind_ch:    tuple of indices (as from np.nonzero())
            indices of the channels that need to be sampled
        segment:    None / int / iterable of ints
            indices of te segments to be sampled
                - None: all
                - int: a single segment
                - iterable of ints: several segments
            Typical usage: None or -1 (last segment)
        radius_max:     None / float
            If provided, only sample the portion of segments that are inside
            the provided ;ajor radius
        concatenate:     bool
            flag indicating whether to concatenate the sampled points per ray
        """

        return _sample.main(
            coll=self,
            key=key,
            key_cam=key_cam,
            res=res,
            mode=mode,
            segment=segment,
            ind_ch=ind_ch,
            radius_max=radius_max,
            concatenate=concatenate,
            return_coords=return_coords,
        )

    # --------------
    # tangency radius
    # --------------

    def get_rays_quantity(
        self,
        key=None,
        key_cam=None,
        quantity=None,
        # limits
        segment=None,
        lim_to_segments=None,
        # for tangency radius
        axis_pt=None,
        axis_vect=None,
    ):
        """ Return a ray-specific quantity of each ray segment

        parameters
        ----------
        quantity:   str
            - 'length': length of rays
            - 'tangency radius': the tangency radius to an axis
        axis_pt:    len=3 iterable
            (x, y, z) coordinates of a pt on the axis, default to [0, 0, 0]
        axis_vect:  len=3 iterable
            (x, y, z) coordinates of the axis vector, default to [0, 0, 1]
        lim_to_segments: bool
            flag indicating whether to limit solutions to the segments

        Return
        -------
        radius:     np.ndarray of floats
            the tangency radii
        kk:         np.ndarray of floats
            the normalized longitudinal coordinate of the tangency points
        ref:        tuple
            The ref tuple on which the data depends
        """

        return _tangency_radius._tangency_radius(
            coll=self,
            key=key,
            key_cam=key_cam,
            quantity=quantity,
            # limits
            segment=segment,
            lim_to_segments=lim_to_segments,
            # for tangency radius
            axis_pt=axis_pt,
            axis_vect=axis_vect,
        )

    def get_rays_intersect_radius(
        self,
        key=None,
        key_cam=None,
        axis_pt=None,
        axis_vect=None,
        axis_radius=None,
        segment=None,
        lim_to_segments=None,
        return_pts=None,
        return_itot=None,
    ):
        """ Return the tangency radius to an axis of each ray segment

        parameters
        ----------
        axis_pt:    len=3 iterable
            (x, y, z) coordinates of a pt on the axis, default to [0, 0, 0]
        axis_vect:  len=3 iterable
            (x, y, z) coordinates of the axis vector, default to [0, 0, 1]
        axis_radius:    float
            The radius around the axis defining the cylinder to intersect
        lim_to_segments: bool
            flag indicating whether to limit solutions to the segments
        return_pts:
            flag indicating whether to return the pts (x, y, z) coordinates

        Return
        -------
        k0:         np.ndarray of floats
            First solution, per segment
        k1:         np.ndarray of floats
            Second solution, per segment
        iok:        np.ndarray of bool
            Flag indicating which segments have at least an intersection
        pts_x:      np.ndarray of floats
            The x coordinates of the points inside the cylinder
        pts_y:      np.ndarray of floats
            The y coordinates of the points inside the cylinder
        pts_z:      np.ndarray of floats
            The z coordinates of the points inside the cylinder

        """

        return _tangency_radius.intersect_radius(
            coll=self,
            key=key,
            key_cam=key_cam,
            axis_pt=axis_pt,
            axis_vect=axis_vect,
            axis_radius=axis_radius,
            segment=segment,
            lim_to_segments=lim_to_segments,
            return_pts=return_pts,
            return_itot=return_itot,
            )

    # --------------
    # Single point camera
    # --------------

    def add_single_point_camera2d(
        self,
        key=None,
        # from rays
        key_rays=None,
        segment=None,
        # user-defined
        cent=None,
        nin=None,
        e0=None,
        e1=None,
        # angles
        angle0=None,
        angle1=None,
        config=None,
        strict=None,
        # optional naming
        key_angle0=None,
        key_angle1=None,
        ref_angle0=None,
        ref_angle1=None,
        units_angles=None,
    ):
        """ Add a set of 2d rays from a single point

        Rays are sampling a portion of sphere around cent

        cent can be either:
            - user-defined
            - derived from intersection of existing rays

        Unit vectors (nin) is also either:
            - derived from existing rays
            - user-defined

        Portion is defined by 2 angles:
            - angle0
            - angle1

        """

        return _single_point_cam.main(
            coll=self,
            key=key,
            # from rays
            key_rays=key_rays,
            segment=segment,
            # user-defined
            cent=cent,
            nin=nin,
            e0=e0,
            e1=e1,
            # angles
            angle0=angle0,
            angle1=angle1,
            config=config,
            strict=strict,
            # optional naming
            key_angle0=key_angle0,
            key_angle1=key_angle1,
            ref_angle0=ref_angle0,
            ref_angle1=ref_angle1,
            units_angles=units_angles,
        )

    # ----------------------
    # get angles of rasy from single point camera
    # ----------------------

    def get_rays_angles_from_single_point_camera2d(
        self,
        key_single_pt_cam=None,
        # rays to get angles of
        key_rays=None,
        segment=None,
        # max tolerance
        tol_radius=None,
        # optional indices and convex
        return_indices=None,
        convex_axis=None,
        # verb
        verb=None,
    ):
        """ Compute (angle0, angle1) for chosen rays

        As seen from a single point camera2d

        Optionally disable rays with an impact factor too large vs camera

        """
        return _single_point_cam._get_rays_angles(
            coll=self,
            key_single_pt_cam=key_single_pt_cam,
            # rays to get angles of
            key_rays=key_rays,
            segment=segment,
            # max tolerance
            tol_radius=tol_radius,
            # optional indices and convex
            return_indices=return_indices,
            convex_axis=convex_axis,
            # verb
            verb=verb,
        )

    # ------------------
    # plotting - static
    # ------------------

    def plot_rays_static(
        self,
        key=None,
        proj=None,
        concatenate=None,
        mode=None,
        res=None,
        # config
        plot_config=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):
        return _plot._plot_rays(
            coll=self,
            key=key,
            proj=proj,
            concatenate=concatenate,
            mode=mode,
            res=res,
            # config
            plot_config=plot_config,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )

    # ------------------
    # plotting - interactive
    # ------------------

    def plot_rays(
        self,
        # keys
        key=None,
        keyZ=None,
        # options
        proj=None,
        res=None,
        # data plot
        data=None,
        units=None,
        cmap=None,
        vmin=None,
        vmax=None,
        alpha=None,
        dx0=None,
        dx1=None,
        # config
        plot_config=None,
        plot_colorbar=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):

        return _class8_plot._plot_diagnostic(
            coll=self,
            isray=True,
            # keys
            key=key,
            keyZ=keyZ,
            # options
            proj=proj,
            los_res=res,
            # data plot
            data=data,
            units=units,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            alpha=alpha,
            dx0=dx0,
            dx1=dx1,
            # config
            plot_config=plot_config,
            plot_colorbar=plot_colorbar,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )

    # --------------
    # sinogram
    # --------------

    def get_sinogram(
        self,
        key=None,
        segment=None,
        # config
        config=None,
        kVes=None,
        # sinogram ref
        R0=None,
        Z0=None,
        # sinogram options
        ang=None,
        ang_units=None,
        impact_pos=None,
        pmax=None,
        Dphimax=None,
        # plotting options
        plot=None,
        sketch=None,
        color=None,
        marker=None,
        label=None,
        # other options
        verb=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
    ):
        """ plot the sinogram of a set of rays / diagnostic LOS

        Optionally plot also the configuration

        The sinogram can be plotted with:
            * ang = 'ksi' or 'theta'
            * ang_units = 'rad' or 'deg'
            * impact_pos = True or False
            * sketch = True or False

        """

        return _sinogram.sinogram(
            coll=self,
            key=key,
            segment=segment,
            # config
            config=config,
            kVes=kVes,
            # sinogram ref
            R0=R0,
            Z0=Z0,
            # sinogram options
            ang=ang,
            ang_units=ang_units,
            impact_pos=impact_pos,
            pmax=pmax,
            Dphimax=Dphimax,
            # plotting options
            plot=plot,
            sketch=sketch,
            color=color,
            marker=marker,
            label=label,
            # other options
            verb=verb,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
        )

    # --------------
    # save to step file (CAD)
    # --------------

    def save_rays_to_stp(
        # ---------------
        # input from tofu
        self,
        key=None,
        key_cam=None,
        # input from arrays
        ptsx=None,
        ptsy=None,
        ptsz=None,
        # input from file
        pfe_in=None,
        # ---------------
        # options
        outline_only=None,
        include_centroid=None,
        factor=None,
        color=None,
        color_by_pixel=None,
        chain=None,
        # ---------------
        # saving
        pfe_save=None,
        overwrite=None,
    ):
        """ Save a set of 'rays' to a stp file (CAD-readable)

        Parameters
        ----------
        key : str, optional
            key to the existing set of rays
        pfe : str, optional
            valid path-file-extension (file str) where to save the stp file

        Returns
        -------
        None.

        """

        return _save2stp.main(
            # ---------------
            # input from tofu
            coll=self,
            key=key,
            key_cam=key_cam,
            # input from arrays
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            # input from file
            pfe_in=pfe_in,
            # ---------------
            # options
            outline_only=outline_only,
            include_centroid=include_centroid,
            factor=factor,
            color=color,
            color_by_pixel=color_by_pixel,
            chain=chain,
            # ---------------
            # saving
            pfe_save=pfe_save,
            overwrite=overwrite,
        )




"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import sys
import warnings
import itertools as itt

# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu as tf
import tofu.spectro as tfs


from .test_data._spectral_constraints import _DLINES_TOT, _DLINES
from .test_data._spectral_constraints import _DCONSTRAINTS, _DCONSTANTS
from .test_data._spectral_constraints import _DOMAIN, _FOCUS, _BINNING
from .test_data._spectral_constraints import _DSCALES, _DX0, _DBOUNDS


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.spectro.fit12d'
_PATH_TEST_DATA = os.path.join(_here, 'test_data')


# #############################################################################
#                           Default values
# #############################################################################


_DLINES_LOCAL = {
    'a': {
        'lambda0': 3.95e-10,
        'delta': 0.001e-10,
        'sigma': 0.002e-10,
        'amp': 1.,
        'noise': 0.01,
        'group': 0,
    },
    'b': {
        'lambda0': 3.97e-10,
        'delta': -0.001e-10,
        'sigma': 0.0015e-10,
        'amp': 0.5,
        'noise': 0.001,
        'group': 1,
    },
    'c': {
        'lambda0': 3.975e-10,
        'delta': 0.001e-10,
        'sigma': 0.001e-10,
        'amp': 0.6,
        'noise': 0.005,
        'group': 0,
    },
    'd': {
        'lambda0': 3.99e-10,
        'delta': 0.002e-10,
        'sigma': 0.002e-10,
        'amp': 0.8,
        'noise': 0.01,
        'group': 1,
    },
}


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


#######################################################
#
#        saving input spect
#       (To be run manually)
#######################################################


def _save_test_spect(path=_PATH_TEST_DATA, dlines=_DLINES_LOCAL, save=True):

    nlamb = 100
    lamb = np.linspace(3.94, 4, nlamb)*1e-10
    var = np.linspace(-25, 25, 51)
    dvar = var[-1] - var[0]

    def bck(lamb=None, offset=None, slope=None):
        return offset + (lamb-lamb.min())*slope/(lamb.max()-lamb.min())

    def gauss(lamb=None, lamb0=None, sigma=None, delta=None, amp=None):
        return amp * np.exp(-(lamb-lamb0-delta)**2/(2*sigma**2))

    def noise(lamb=None, amp=None, freq=None, phase=None):
        return amp*np.sin(
            lamb*freq*2.*np.pi/(lamb.max()-lamb.min()) + phase
        )

    spect2d = bck(
        lamb=lamb[None, :],
        offset=0.1*np.exp(-(var[:, None]-25)**2/10**2),
        slope=0.001,
    )
    spect2d += noise(lamb=lamb[None, :], amp=0.01, freq=10, phase=0.)

    for ii, k0 in enumerate(dlines.keys()):
        spect2d += gauss(
            lamb=lamb[None, :],
            amp=dlines[k0]['amp'] * np.exp(-var[:, None]**2/20**2),
            lamb0=dlines[k0]['lambda0'],
            sigma=dlines[k0]['sigma']*(
                1 + 2*(ii/len(dlines))*np.cos(var[:, None]*2*np.pi/dvar)
            ),
            delta=dlines[k0]['delta']*(
                1 + 2*(ii/len(dlines))*np.sin(
                    var[:, None]*2*np.pi*(len(dlines)-ii)/dvar
                )
            ),
        )
        spect2d += noise(
            lamb=lamb[None, :],
            amp=dlines[k0]['noise'] * np.exp(-var[:, None]**2/10**2),
            freq=10*(len(dlines)-ii),
            phase=ii,
        )

    # --------
    # save
    if save:
        pfe = os.path.join(path, 'spectral_fit.npz')
        np.savez(pfe, lamb=lamb, var=var, dlines=dlines, spect2d=spect2d)


def _load_test_spect(path=_PATH_TEST_DATA):
    pfe = os.path.join(path, 'spectral_fit.npz')
    out = dict(np.load(pfe, allow_pickle=True))
    return out['lamb'], out['var'], out['dlines'].tolist(), out['spect2d']


# #############################################################################
#
#     Creating Ves objects and testing methods
#
# #############################################################################


class Test01_ProofOfPrinciple(object):

    @classmethod
    def setup_class(cls):

        # load test data
        lamb, var, dlines, spect2d = _load_test_spect()
        mask = np.repeat((np.abs(var-15) > 3)[:, None], lamb.size, axis=1)

        # Plot spect 2d
        # fig = plt.figure(figsize=(12, 10));
        # ax0 = fig.add_axes([0.05, 0.1, 0.4, 0.8])
        # ax0.set_xlabel(r'$\lambda$ (m)')
        # ax0.set_ylabel(r'$y$ (m)')
        # extent = (lamb.min(), lamb.max(), var.min(), var.max())
        # ax0.imshow(spect2d, extent=extent, aspect='auto', origin='lower');
        # sp2bis = np.copy(spect2d)
        # sp2bis[mask] = np.nan
        # ax1 = fig.add_axes([0.55, 0.1, 0.4, 0.8])
        # ax1.set_xlabel(r'$\lambda$ (m)')
        # ax1.set_ylabel(r'$y$ (m)')
        # ax1.imshow(sp2bis, extent=extent, aspect='auto', origin='lower');
        # plt.ion()
        # plt.show();
        # import pdb; pdb.set_trace()     # DB

        # Define constraint dict
        defconst = {
            'amp': False,
            'width': False,
            'shift': False,
            'double': False,
            'symmetry': False,
        }

        ldconst = [
            {
                'amp': {'a1': ['a', 'd']},
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'b': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's3', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': True,
                'symmetry': True,
            },
            {
                'amp': False,
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's2', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': False,
                'symmetry': False,
            },
        ]

        ldx0 = [
            None,
            {
                # 'amp': {''},
                'width': 1.,
                'shift': {
                    's2': 1.,
                },
                'dratio': 0,
                'dshift': 0,
            }
        ]

        ldomain = [
            None,
            {
                'lamb': [
                    [3.94e-10, 3.952e-10],
                    (3.95e-10, 3.956e-10),
                    [3.96e-10, 4e-10],
                ],
            },
        ]

        ldata = [
            spect2d[5, :],
            spect2d[5:8, :],
        ]

        lfocus = [None, 'b', [3.94e-10, 3.96e-10]]

        ldconstants = [None, {'shift': {'s2': 0}}]

        cls.lamb = lamb
        cls.var = var
        cls.dlines = dlines
        cls.spect2d = spect2d
        cls.mask = mask
        cls.defconst = defconst
        cls.ldconst = ldconst
        cls.ldx0 = ldx0
        cls.ldomain = ldomain
        cls.ldata = ldata
        cls.lfocus = lfocus
        cls.ldconstants = ldconstants
        cls.ldinput1d = []
        cls.ldinput1d_run = []
        cls.ldfit1d = []
        cls.ldex1d = []
        cls.ldinput2d = []
        cls.ldinput2d_run = []
        cls.ldfit2d = []
        cls.ldex2d = []

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_fit1d_dinput(self, full=None):
        """ Build the input dict for fitting a 1d spectrum
        """

        combin = [
            self.ldconst, self.ldx0, self.ldomain, self.ldata,
            self.lfocus, self.ldconstants,
        ]
        nn = int(np.prod([len(cc) for cc in combin]))
        for ii, comb in enumerate(itt.product(*combin)):

            pos = ii % 2 == 0
            dinput = tfs.fit1d_dinput(
                dlines=self.dlines,
                dconstraints=comb[0],
                dconstants=comb[5],
                dprepare=None,
                data=np.copy(comb[3]),
                lamb=self.lamb,
                mask=None,
                domain=comb[2],
                pos=pos,
                subset=None,
                same_spectrum=None,
                nspect=None,
                same_spectrum_dlamb=None,
                focus=comb[4],
                valid_fraction=0.28,     # fraction of pixels ok per time step
                valid_nsigma=0,         # S/N ratio for each pixel
                focus_half_width=None,
                valid_return_fract=None,
                dscales=None,
                dx0=comb[1],
                dbounds=None,
                defconst=self.defconst,
            )
            self.ldinput1d.append(dinput)

            # run (fit1d) only for some cases
            run = False
            if comb[1] is None and comb[5] is not None and comb[4] != 'b':
                run = True
            self.ldinput1d_run.append(run)

    def test02_funccostjac_1d(self):
        """ check that tofu properly returns 3 functions for fitting 1d spectra

        func_detail: should return all components of a spectrum
        func_cost: shoud return only the total minus the original data
        func_jac: should return the jacobian

        in principle: sum(func_detail) == func_cost(data=0)

        """

        func = tfs._fit12d_funccostjac.multigausfit1d_from_dlines_funccostjac
        for ii, dd in enumerate(self.ldinput1d):
            func_detail, func_cost, func_jac = func(
                lamb=dd['dprepare']['lamb'],
                dinput=dd,
                dind=dd['dind'],
                jac='dense',
            )

            # x0
            x0 = tfs._fit12d_dinput._dict2vector_dscalesx0bounds(
                dd=dd['dx0'], dd_name='dx0', dinput=dd,
            )

            # scales
            scales = tfs._fit12d_dinput._dict2vector_dscalesx0bounds(
                dd=dd['dscales'], dd_name='dscales', dinput=dd,
            )

            # y0
            y0 = func_detail(x0[0, :], scales=scales[0, :])
            y1 = func_cost(
                x0[0, :],
                scales=scales[0, :],
                data=dd['dprepare']['data'][0, :],
            )

            # check consistency between func_detail and func_cost
            assert np.allclose(
                np.sum(y0, axis=1) - dd['dprepare']['data'][0, :],
                y1,
                equal_nan=True,
            )

    def test03_fit1d(self, strict=None, verb=None):
        """ Actually run the 1d spectrum fitting routine,

        """
        if verb:
            nprod = np.array(self.ldinput1d_run).sum()
            nn = len(f'\tspectrum {nprod} / {nprod}')

        for ii, inn in enumerate(np.array(self.ldinput1d_run).nonzero()[0]):

            if verb:
                msg = f"\tspectrum {ii+1} / {nprod}".ljust(nn)
                print(msg, end='\r', flush=True)

            chain = ii % 2 == 0
            dfit1d = tfs.fit1d(
                dinput=self.ldinput1d[inn],
                method=None,
                Ti=None,
                chain=chain,
                jac='dense',
                verbose=False,
                strict=strict,
                plot=False,
            )
            assert np.sum(dfit1d['validity'] < 0) == 0
            self.ldfit1d.append(dfit1d)

    def test04_fit1d_dextract(self):
        """ Extract dict of output from fitted 1d spectra
        """
        for ii, dd in enumerate(self.ldfit1d):
            dex = tfs.fit1d_extract(
                dfit1d=dd,
                ratio=('a', 'c'),
                sol_total=True,
                sol_detail=ii % 2 == 0,
            )
            self.ldex1d.append(dex)

    def test05_fit1d_plot(self, warn=True):
        lwar = []
        for ii, dd in enumerate(self.ldex1d):
            try:
                # For a yet unknown reason, this particular test crashes on
                # Windows only due to figure creation at
                # tfs._plot.plot_fit1d(): line 337
                # already investigated: reducing figure size and early closing
                # No more ideas...
                # This link suggests it may have something to do with 
                # inches => pixels conversion of figure size...
                # https://github.com/matplotlib/matplotlib/issues/14225
                if 'win' not in sys.platform.lower():
                    dax = tfs._plot.plot_fit1d(
                        dfit1d=self.ldfit1d[ii],
                        dextract=dd,
                        annotate=self.ldfit1d[ii]['dinput']['keys'][0],
                        fs=(4, 4),
                    )
            except Exception as err:
                if warn:
                    lwar.append((ii, str(err)))
                else:
                    raise err
            finally:
                plt.close('all')

        if len(lwar) > 0:
            msg = (
                "\nThe ({}/{}) following fit1d plots failed:\n".format(
                    len(lwar), len(self.ldex1d),
                )
                + "\n".join(["\t- {}: {}".format(ww[0], ww[1]) for ww in lwar])
            )
            warnings.warn(msg)

    def test06_fit2d_dinput(self):
        """ Build the input dict for fitting a 2d spectrum
        """
        combin = [
            self.ldconst, self.ldx0, self.ldomain,
            self.lfocus, self.ldconstants,
        ]
        ntot = (
            len(self.ldconst) * len(self.ldx0) * len(self.ldomain)
            * len(self.lfocus) * len(self.ldconstants)
        )
        for ii, comb in enumerate(itt.product(*combin)):

            # additional parameters
            pos = ii % 2 == 0
            mask = self.mask if ii % 3 == 0 else None
            binning = False if ii % 2 == 0 else {'lamb': 85, 'phi': 40}

            # set dinput
            dinput = tfs.fit2d_dinput(
                dlines=self.dlines,
                dconstraints=comb[0],
                dconstants=comb[4],
                dprepare=None,
                data=np.copy(self.spect2d),
                lamb=self.lamb,
                phi=self.var,
                mask=mask,
                nbsplines=5,
                domain=comb[2],
                pos=pos,
                subset=None,
                binning=binning,
                focus=comb[3],
                valid_fraction=0.28,     # fraction of pixels ok per time step
                valid_nsigma=0.2,         # S/N ratio for each pixel
                focus_half_width=None,
                valid_return_fract=None,
                dscales=None,
                dx0=comb[1],
                dbounds=None,
                defconst=self.defconst,
            )
            self.ldinput2d.append(dinput)

            c0 = (
                comb[1] == self.ldx0[0]
                and comb[2] == self.ldomain[1]
                and comb[3] == self.lfocus[0]
                and comb[4] == self.ldconstants[0]
            )
            if c0:
                self.ldinput2d_run.append(ii)

    def test07_plot_dinput2d(self):
        for ii, dd in enumerate(self.ldinput2d):
            dax = tfs._plot.plot_dinput2d(dinput=dd)
        plt.close('all')

    def test08_funccostjac_2d(self):
        """ check that tofu properly returns 3 functions for fitting 1d spectra

        func_detail: should return all components of a spectrum
        func_sum: shoud return the total spectrum
        func_cost: shoud return the total spectrum minus the original data
        func_jac: should return the jacobian

        in principle: sum(func_detail()) == func_sum() == func_cost(data=0)

        """
        func = tfs._fit12d_funccostjac.multigausfit2d_from_dlines_funccostjac
        for ii, dd in enumerate(self.ldinput2d):

            if ii < 24:
                continue

            lamb_flat = dd['dprepare']['lamb'].ravel()
            phi_flat = dd['dprepare']['phi'].ravel()
            data_flat = dd['dprepare']['data'][0, ...].ravel()
            func_detail, func_sum, func_cost, func_jac = func(
                phi_flat=phi_flat,
                dinput=dd,
                dind=dd['dind'],
                jac='dense',
            )

            # x0
            x0 = tfs._fit12d_dinput._dict2vector_dscalesx0bounds(
                dd=dd['dx0'], dd_name='dx0', dinput=dd,
            )

            # scales
            scales = tfs._fit12d_dinput._dict2vector_dscalesx0bounds(
                dd=dd['dscales'], dd_name='dscales', dinput=dd,
            )

            # lambrel qnd lambn
            lambrel_flat = lamb_flat - dd['lambmin_bck']
            lambn_flat = lamb_flat[..., None] / dd['lines'][None, :]

            # dy0 vs dy1
            y0 = func_detail(
                x0[0, :],
                lamb=lamb_flat,
                phi=phi_flat,
                scales=scales[0, :],
            )
            indnan = np.all(np.all(np.isnan(y0), axis=-1), axis=-1)
            dy0 = np.nansum(np.nansum(y0, axis=-1), axis=-1) - data_flat
            # beware, nansum returns 0 for all-nans in numpy >= 1.9.0
            dy0[indnan] = np.nan
            dy1 = func_sum(
                x0[0, :],
                lamb=lamb_flat,
                phi=phi_flat,
                scales=scales[0, :],
            ) - data_flat
            dy2 = func_cost(
                x0[0, :],
                scales=scales[0, :],
                phi_flat=phi_flat,
                lambrel_flat=lambrel_flat,
                lambn_flat=lambn_flat,
                data_flat=data_flat,
                indok_flat=np.ones((phi_flat.size,), dtype=bool)
            )
            # check consistency between func_detail and func_cost
            assert np.sum(np.isfinite(dy0)) == np.sum(np.isfinite(dy1))
            assert np.sum(np.isfinite(dy0)) == np.sum(np.isfinite(dy2))
            assert np.allclose(dy0, dy1, equal_nan=True)
            assert np.allclose(dy0, dy2, equal_nan=True)

    def test09_fit2d(self, strict=None, verb=False):
        """ Actually run the 2d spectrum fitting routine,

        """
        for ii, ij in enumerate(self.ldinput2d_run):
            din = self.ldinput2d[ij]
            chain = ii % 2 == 0

            dfit2d = tfs.fit2d(
                dinput=din,
                method=None,
                Ti=None,
                chain=chain,
                jac='dense',
                verbose=verb,
                strict=strict,
                plot=False,
            )
            assert np.sum(dfit2d['validity'] < 0) == 0
            self.ldfit2d.append(dfit2d)

    def test09_fit2d_dextract(self):
        """ Extract dict of output from fitted 2d spectra
        """
        for ii, dd in enumerate(self.ldfit2d):
            dex = tfs.fit2d_extract(
                dfit2d=dd,
                ratio=('a', 'c'),
                sol_total=True,
                sol_detail=ii % 2 == 0,
            )
            self.ldex2d.append(dex)

    def test10_fit2d_plot(self, warn=True):
        lwar = []
        for ii, dd in enumerate(self.ldex2d):
            try:
                # For a yet unknown reason, this particular test crashes on
                # Windows only due to figure creation at
                # tfs._plot.plot_fit1d(): line 337
                # already investigated: reducing figure size and early closing
                # No more ideas...
                # This link suggests it may have something to do with
                # inches => pixels conversion of figure size...
                # https://github.com/matplotlib/matplotlib/issues/14225
                if 'win' not in sys.platform.lower():
                    dax = tfs._plot.plot_fit2d(
                        dfit2d=self.ldfit2d[ii],
                        dextract=dd,
                        annotate=self.ldfit2d[ii]['dinput']['keys'][0],
                        fs=(4, 4),
                    )
            except Exception as err:
                if warn:
                    lwar.append((ii, str(err)))
                else:
                    raise err
            finally:
                plt.close('all')


class Test02_RealisticWESTCase(object):

    def test00_load_dinput_dfit_dextract(
        self,
        verb=None,
        # non-default parameters
        symmetry=None,
        mask=None,
        domain=None,
        dconstants=None,
        binning=None,
        dscales=None,
        dx0=None,
        dbounds=None,
        focus=None,
        valid_fraction=None,
        valid_nsigma=None,
        nbsplines=None,
        tol=None,
        # debug
        plot_dinput=False,
        vmin=None,
        vmax=None,
        # extract
        sol_detail=None,
        phi_prof=None,
        phi_npts=None,
    ):

        # load test data
        shot = 54046
        dout = dict(np.load(
            os.path.join(_PATH_TEST_DATA, f'west_{shot}_xics.npz'),
            allow_pickle=True,
        ))

        lk = ['data', 't', 'indt', 'nt', 'names', 'dunits', 'dbonus']
        data, t, indt, nt, names, dunits, dbonus = [dout[k0] for k0 in lk]
        dunits = dunits.tolist()
        dbonus = dbonus.tolist()

        # reshape data into a (nt, nxi, nxj) array
        nt = t.size
        nxi, nxj = dbonus['nH'], dbonus['nV']
        data = data.reshape((nt, nxj, nxi)).swapaxes(1, 2)

        # ----------------
        # Get crystal and det

        det = dict(np.load(
            os.path.join(_PATH_TEST_DATA, 'det37_CTVD_incC4_New.npz'),
            allow_pickle=True,
        ))

        cryst = tf.load(os.path.join(
            _PATH_TEST_DATA,
            'TFG_CrystalBragg_ExpWEST_DgXICS_ArXVII_sh00000_Vers1.5.0.npz',
        ))
        angle = 1.3124
        # tlim2 = [32, 46]
        cryst.move(angle*np.pi/180.)

        # get lamb / phi
        xi = 172e-6*(np.arange(0, 487) - (487 - 1)/2.)
        xj = 172e-6*(np.arange(0, 1467) - (1467 - 1)/2.)
        bragg, phi, lamb = cryst.get_lambbraggphi_from_ptsxixj_dthetapsi(
            det=det, xi=xi, xj=xj, grid=True,
        )

        # computed from a unique point on the crystal (the summit)
        lamb, bragg, phi = lamb[..., 0], bragg[..., 0], phi[..., 0]

        # ------------------------------------
        # Start data treatment: dict of inputs

        # get dict of constraints
        (
            dlines, dconstraints, dconstants, domain, focus, defconst,
            valid_fraction, valid_nsigma, binning,
            dscales, dx0, dbounds,
            nbsplines, deg, mask, tol,
        ) = get_constraints(
            shot=shot,
            cryst=cryst,
            det=det,
            # non-default parameters
            symmetry=symmetry,
            mask=mask,
            domain=domain,
            dconstants=dconstants,
            binning=binning,
            dscales=dscales,
            dx0=dx0,
            dbounds=dbounds,
            focus=focus,
            valid_fraction=valid_fraction,
            valid_nsigma=valid_nsigma,
            nbsplines=nbsplines,
            tol=tol,
        )

        # format into a unique input dict
        dinput = tf.spectro.fit2d_dinput(
            dlines=dlines,
            dconstraints=dconstraints,
            dconstants=dconstants,
            dprepare=None,
            deg=deg,
            nbsplines=nbsplines,
            knots=None,
            data=data,
            lamb=lamb,
            phi=phi,
            mask=mask,
            domain=domain,
            pos=None,
            subset=None,
            binning=binning,
            cent_fraction=None,
            focus=focus,
            valid_fraction=valid_fraction,
            valid_nsigma=valid_nsigma,
            focus_half_width=None,
            valid_return_fract=None,
            dscales=dscales,
            dx0=dx0,
            dbounds=dbounds,
            nxi=nxi,
            nxj=nxj,
            lphi=None,
            lphi_tol=None,
            defconst=defconst,
        )

        # --------------
        # Optional debug

        if plot_dinput is not False:
            ldax = tf.spectro._plot.plot_dinput2d(
                dinput=dinput,
                indspect=plot_dinput,
                vmin=vmin,
                vmax=vmax,
            )
            return dinput, ldax

        # ------------------------------------
        # Perform main data treatment: fit

        dfit2d = tf.spectro.fit2d(
            dinput=dinput,
            method=None,
            tr_solver=None,
            tr_options=None,
            xtol=tol,
            ftol=tol,
            gtol=tol,
            max_nfev=None,
            loss=None,
            chain=True,
            jac=None,
            verbose=verb,
            strict=True,    # to make sure a corrupted time step doesn't stop
            save=False,
            plot=False,
        )

        # ------------------------------------
        # Finish data treatment: extract quantities

        dextract = tf.spectro.fit2d_extract(
            dfit2d=dfit2d,
            bck=True,
            amp=True,
            ratio=[('ArXVII_w_Bruhns', 'ArXVI_k_Adhoc200408')],
            Ti=True,
            width=True,
            vi=True,
            shift=True,
            sol_total=True,
            sol_detail=sol_detail,
            sol_lamb_phi=None,
            phi_prof=phi_prof,
            phi_npts=phi_npts,
            vs_nbs=None,
        )

        return dinput, dfit2d, dextract, t


def get_constraints(
    shot=None,
    cryst=None,
    det=None,
    # non-default parameters
    symmetry=None,
    mask=None,
    domain=None,
    dconstants=None,
    binning=None,
    dscales=None,
    dx0=None,
    dbounds=None,
    focus=None,
    valid_fraction=None,
    valid_nsigma=None,
    nbsplines=None,
    tol=None,
):

    crystname = 'ArXVII'

    # dlines
    dlines = {
        k0: dict(_DLINES_TOT[k0])
        for k0 in _DLINES[crystname]
    }

    # dconstraints
    dconstraints = dict(_DCONSTRAINTS[crystname])
    if symmetry is not None:
        dconstraints['symmetry'] = symmetry

    # dconstants
    if dconstants is False:
        dconstants = None
    elif dconstants is None:
        dconstants = dict(_DCONSTANTS[crystname])

    # domain
    if domain is False:
        domain = None
    elif domain is None:
        if _DOMAIN[crystname] is not None:
            domain = dict(_DOMAIN[crystname])

    # binning
    if binning is False:
        binning = None
    elif binning is None:
        if _BINNING[crystname] is not None:
            binning = dict(_BINNING[crystname])

    # dscales
    if dscales is False:
        dscales = None
    elif dscales is None:
        if _DSCALES[crystname] is not None:
            dscales = dict(_DSCALES[crystname])

    # dx0
    if dx0 is False:
        dx0 = None
    elif dx0 is None:
        if _DX0[crystname] is not None:
            dx0 = dict(_DX0[crystname])

    # dbounds
    if dbounds is False:
        dbounds = False
    elif dbounds is None:
        if _DBOUNDS[crystname] is not None:
            dbounds = dict(_DBOUNDS[crystname])

    # focus
    if focus is False:
        focus = None
    elif focus is None:
        if _FOCUS[crystname] is not None:
            focus = _FOCUS[crystname]

    # valid
    if valid_fraction is None:
        valid_fraction = 0.6
    if valid_nsigma is None:
        valid_nsigma = 4

    # nbsplines
    if nbsplines is None:
        nbsplines = 15

    # defconst
    defconst = {
        'bck_amp': False, 'bck_rate': False, 'amp': False,
        'width': False, 'shift': False, 'double': False,
    }

    # deg
    deg = 2

    # mask
    # (nxj, nxi) array
    # False where corrupt pixels
    if mask is None:
        pfe = os.path.join(_PATH_TEST_DATA, '_mask_54041.npz')
        mask = dict(np.load(pfe, allow_pickle=True))
        mask = ~np.any(mask['ind'], axis=0).T

    # tol
    if tol is None:
        tol = 1.e-2

    return (
        dlines, dconstraints, dconstants, domain, focus, defconst,
        valid_fraction, valid_nsigma, binning,
        dscales, dx0, dbounds,
        nbsplines, deg, mask, tol,
    )



# -*- coding: utf-8 -*-


import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp
import datastock as ds


# ###############################################################
# ###############################################################
#                   main
# ###############################################################


def main(
    coll=None,
    key=None,
    key_cam=None,
    res=None,
    mode=None,
    segment=None,
    ind_ch=None,
    radius_max=None,
    concatenate=None,
    return_coords=None,
):

    # ------------
    # check inputs

    (
        res, mode, concatenate,
        segment, radius_max,
        return_coords, out_xyz, out_k, out_l,
    ) = _sample_check(
        res=res,
        mode=mode,
        concatenate=concatenate,
        segment=segment,
        radius_max=radius_max,
        return_coords=return_coords,
    )

    # -----------
    # prepare
    # -----------

    # ------------------------
    # get points of interest

    # get rays points (nseg+1, nx0, nx1, ...)
    pts_x, pts_y, pts_z = coll.get_rays_pts(key=key, key_cam=key_cam)

    # extract sizes
    npts = pts_x.shape[0]
    i0 = np.arange(0, npts)

    # ---------------------------
    # optional segment selection

    if segment is not None:
        segment[segment < 0] = npts - 1 + segment[segment < 0]
        iseg = np.r_[segment, segment[-1] + 1]

    # ---------------------------------------
    # optional ind_flat (selection of pixels)

    if ind_ch is not None:

        # check ind
        ind_ch = _check_ind_channels(ind_ch, shape=pts_x.shape[1:], key=key)

        sli = tuple([slice(None)] + list(ind_ch))
        pts_x = pts_x[sli]
        pts_y = pts_y[sli]
        pts_z = pts_z[sli]

    # -------------------------------
    # get original length per segment

    length_orig = np.sqrt(
        np.diff(pts_x, axis=0)**2
        + np.diff(pts_y, axis=0)**2
        + np.diff(pts_z, axis=0)**2
    )

    # total length from starting point
    zeros = np.zeros(tuple(np.r_[1, length_orig.shape[1:]]), dtype=float)
    length0 = np.cumsum(np.concatenate((zeros, length_orig), axis=0), axis=0)
    length1 = np.concatenate((length_orig, length_orig[-1:, ...]), axis=0)

    # ---------------------------
    # changes to pts

    if radius_max is not None:
        pts_x, pts_y, pts_z, i0 = coll.get_rays_intersect_radius(
            key=key,
            key_cam=key_cam,
            segment=segment,
            axis_radius=radius_max,
            lim_to_segments=True,
            return_pts=True,
            return_itot=True,
        )[3:]

        # ind_flat
        if ind_ch is not None:
            sli = tuple([slice(None)] + list(ind_ch))
            pts_x = pts_x[sli]
            pts_y = pts_y[sli]
            pts_z = pts_z[sli]
            i0 = i0[sli]

        length_rad = np.sqrt(
            np.diff(pts_x, axis=0)**2
            + np.diff(pts_y, axis=0)**2
            + np.diff(pts_z, axis=0)**2
        )

        if segment is not None:
            length0 = length0[iseg, ...]
            length1 = length1[iseg, ...]

    else:
        if segment is not None:
            i0 = i0[iseg]
            pts_x = pts_x[i0, ...]
            pts_y = pts_y[i0, ...]
            pts_z = pts_z[i0, ...]
            length0 = length0[iseg, ...]
            length1 = length1[iseg, ...]

        length_rad = length0[1:, ...]

    # -----------
    # trivial
    # -----------

    if pts_x.size == 0 or not np.any(np.isfinite(pts_x)):
        return [None]*len(return_coords)

    # --------------------
    # compute sampling
    # ---------------------

    # ------------------
    # relative sampling
    # => each segment / pixel has ame number of points

    # rel
    if mode == 'rel':

        # make sure npts allow to describe all integer indices
        nptsi = (i0[-1] - i0[0]) * int(np.ceil(1./res))
        N = int(np.ceil((nptsi - 1) / (i0[-1] - i0[0])))
        nptsi = N * (i0[-1] - i0[0]) + 1
        itot = np.linspace(i0[0], i0[-1], nptsi)

        # interpolate
        if out_xyz:
            pts_x = scpinterp.interp1d(
                i0,
                pts_x,
                kind='linear',
                axis=0,
            )(itot)
            pts_y = scpinterp.interp1d(
                i0,
                pts_y,
                kind='linear',
                axis=0,
            )(itot)
            pts_z = scpinterp.interp1d(
                i0,
                pts_z,
                kind='linear',
                axis=0,
            )(itot)

        # get length
        if out_l:
            i1 = np.floor(itot).astype(int)
            i1[i1 == length1.shape[0] - 1] -= 1
            length = length1[i1, ...]
            lengthtot = length0[i1, ...]

    # --------------------
    # absolute sampling
    # different nb of pts per segment / pixel

    else:

        # get indices of valid pts
        iok = np.isfinite(pts_x)

        # get shape of arrays based on max nb of points ()
        nn = np.ceil(length_rad / res).astype(int)
        nmax = np.nansum(nn, axis=0) + 1
        shape = tuple(np.r_[np.nanmax(nmax), length_rad.shape[1:]])

        # initialize arrays
        itot = np.full(shape, np.nan)

        if out_xyz:
            lpx = np.full(shape, np.nan)
            lpy = np.full(shape, np.nan)
            lpz = np.full(shape, np.nan)

        if out_l:
            llen = np.full(shape, np.nan)
            lentot = np.full(shape, np.nan)

        # loop in pixels
        for ind in itt.product(*[range(ss) for ss in pts_x.shape[1:]]):

            sli0 = tuple([slice(None)] + list(ind))
            ioki = iok[sli0]
            iokin = ioki[:-1] * ioki[1:]

            # trivial case
            if not np.any(iokin):
                continue

            # get sli2 and i0i
            sli2 = tuple([ioki] + list(ind))
            if radius_max is None:
                i0i = i0[ioki]
            else:
                i0i = i0[sli2]
            nni = nn[tuple([iokin] + list(ind))]

            # itoti
            itoti = []
            anyok = False
            for jj in range(i0i.size - 1):
                if np.all(np.isfinite(i0i[jj:jj+2])):
                    itoti.append(
                        np.linspace(
                            i0i[jj],
                            i0i[jj+1],
                            nni[jj] + 1,
                        )[:-1]
                    )
                    anyok = True

            if anyok is False:
                continue

            if np.isfinite(i0i[-1]):
                itoti.append([i0i[-1]])

            itoti = np.concatenate(tuple(itoti))

            # safety check
            ni = itoti.size
            if nmax[ind] != ni:
                msg = (
                    "Mismatch between:\n"
                    f"\t- nmax[{ind}] = {nmax[ind]}\n"
                    f"\t- itoti.size = {itoti.size}\n"
                )
                raise Exception(msg)

            slin = tuple([np.arange(ni)] + list(ind))
            itot[slin] = itoti

            # interpolate
            if out_xyz:
                lpx[slin] = scpinterp.interp1d(
                    i0i,
                    pts_x[sli2],
                    kind='linear',
                    axis=0,
                )(itoti)

                lpy[slin] = scpinterp.interp1d(
                    i0i,
                    pts_y[sli2],
                    kind='linear',
                    axis=0,
                )(itoti)

                lpz[slin] = scpinterp.interp1d(
                    i0i,
                    pts_z[sli2],
                    kind='linear',
                    axis=0,
                )(itoti)

            if out_l:
                i1 = np.floor(itoti).astype(int)
                i1[i1 == ioki.sum()] -= 1

                llen[slin] = length1[sli2][i1]
                lentot[slin] = length0[sli2][i1]

        if out_xyz:
            pts_x, pts_y, pts_z = lpx, lpy, lpz
        if out_l:
            length = llen
            lengthtot = lentot

    # -------------
    # adjust npts
    # -------------

    iok = np.any(np.isfinite(itot), axis=tuple(range(1, itot.ndim)))
    if not np.all(iok):
        sli = (iok,) + tuple([slice(None) for ii in itot.shape[1:]])

        itot = itot[sli]
        if out_xyz:
            pts_x = pts_x[sli]
            pts_y = pts_y[sli]
            pts_z = pts_z[sli]
        if out_l:
            length = length[sli]
            lengthtot = lengthtot[sli]

    # -------------------------------------
    # optional concatenation (for plotting)
    # -------------------------------------

    if concatenate is True:

        # if mode == 'rel':
        shape_nan = tuple(np.r_[np.r_[1], pts_x.shape[1:]])
        nan = np.full(shape_nan, np.nan)

        if out_k:
            itot2 = np.full(pts_x.shape, np.nan)
            for ii in range(pts_x.shape[0]):
                itot2[ii, ...] = itot[ii]
            itot = remove_consecutive_nans(
                np.concatenate((itot2, nan), axis=0).T.ravel()
            )

        if out_xyz:
            pts_x = remove_consecutive_nans(
                np.concatenate((pts_x, nan), axis=0).T.ravel()
            )
            pts_y = remove_consecutive_nans(
                np.concatenate((pts_y, nan), axis=0).T.ravel()
            )
            pts_z = remove_consecutive_nans(
                np.concatenate((pts_z, nan), axis=0).T.ravel()
            )

        if out_l:
            length = remove_consecutive_nans(
                np.concatenate((length, nan), axis=0).T.ravel()
            )
            lengthtot = remove_consecutive_nans(
                np.concatenate((lengthtot, nan), axis=0).T.ravel()
            )

    # -------------
    # adjust kk
    # -------------

    if out_k:
        kk = itot - np.floor(itot)
        kk[itot == np.nanmax(itot)] = 1.

    # -------------
    # return
    # -------------

    lout = []
    for cc in return_coords:
        if cc == 'x':
            lout.append(pts_x)

        elif cc == 'y':
            lout.append(pts_y)

        elif cc == 'z':
            lout.append(pts_z)

        elif cc == 'R':
            lout.append(np.hypot(pts_x, pts_y))

        elif cc == 'phi':
            lout.append(np.arctan2(pts_y, pts_x))

        elif cc == 'ang_vs_ephi':
            phi = np.arctan2(pts_y, pts_x)
            ux = np.diff(pts_x, axis=0)
            uy = np.diff(pts_y, axis=0)
            ux = np.concatenate((ux[0:1, ...], ux), axis=0)
            uy = np.concatenate((uy[0:1, ...], uy), axis=0)
            vn = np.sqrt(ux**2 + uy**2)
            ux = ux / vn
            uy = uy / vn
            lout.append(np.arccos(-np.sin(phi)*ux + np.cos(phi)*uy))

        elif cc == 'itot':
            lout.append(itot)

        elif cc == 'k':
            lout.append(kk)

        elif cc == 'l':
            lout.append(kk*length)

        elif cc == 'ltot':
            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.subplot(1,4,1)
            # plt.plot(kk)
            # plt.subplot(1,4,2)
            # plt.plot(length)
            # plt.subplot(1,4,3)
            # plt.plot(lengthtot)
            # plt.subplot(1,4,4)
            # plt.plot(kk*length + lengthtot)
            # import pdb; pdb.set_trace()     # DB

            lout.append(kk*length + lengthtot)

    return lout


# ###############################################################
#                   check inputs
# ###############################################################


def _sample_check(
    res=None,
    mode=None,
    segment=None,
    radius_max=None,
    concatenate=None,
    return_coords=None,
):

    # -------------
    # booleans
    # -------------

    # ---------
    # res

    res = ds._generic_check._check_var(
        res, 'res',
        types=float,
        default=0.1,
        sign='> 0',
    )

    # ---------
    # mode

    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='rel',
        allowed=['rel', 'abs'],
    )

    # ------------
    # concatenate

    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=False,
    )

    # ----------------
    # segment indices
    # ----------------

    if segment is not None:
        segment = np.atleast_1d(segment).astype(int).ravel()

    # --------------------
    # tangency_radius_max

    if radius_max is not None:
        radius_max = ds._generic_check._check_var(
            radius_max, 'radius_max',
            types=(float, int),
            sign='> 0',
        )

        if mode == 'rel':
            msg = "radius_max can only be used with mode='abs'!"
            raise Exception(msg)

    # -------------
    # return_coords
    # -------------

    if isinstance(return_coords, str):
        return_coords = [return_coords]

    lok_k = ['k', 'l', 'ltot', 'itot']
    lok_xyz = ['x', 'y', 'z', 'R', 'phi', 'ang_vs_ephi']
    return_coords = ds._generic_check._check_var_iter(
        return_coords, 'return_coords',
        types=list,
        types_iter=str,
        default=['x', 'y', 'z'],
        allowed=lok_k + lok_xyz,
    )

    out_xyz = any([ss in return_coords for ss in lok_xyz])
    out_k = any([ss in return_coords for ss in lok_k])
    out_l = any(['l' in return_coords, 'ltot' in return_coords])

    return (
        res, mode, concatenate,
        segment, radius_max,
        return_coords, out_xyz, out_k, out_l
    )


# ###############################################################
#                   check indices of channels
# ###############################################################


def _check_ind_channels(
    ind_ch=None,
    shape=None,
    key=None,
):
    """ Make sure ind_ch is a tuple (one dim per dim of shape)

    Make sure it is made of slice or np.ndarray
    To preserve the dimensions of the array even if ind_ch is a scalar

    Parameters
    ----------
    ind_ch : tuple of np.ndarrays, like from np.nonzero() or np.unravel_index()
        indices of the desired channels
    shape : TYPE, optional
        DESCRIPTION. The default is None.
    key : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    ind_ch : TYPE
        DESCRIPTION.

    """

    # ------------------------
    # special retro-compatible case
    # ------------------------

    if np.isscalar(ind_ch) and len(shape) > 1:
        ind_ch = np.unravel_index(ind_ch, shape)

    # ------------------------
    # basic conformity checks
    # ------------------------

    # single index
    if len(shape) == 1 and not isinstance(ind_ch, tuple):
        ind_ch = (ind_ch,)

    # check is tuple
    c0 = (
        isinstance(ind_ch, tuple)
        and len(ind_ch) == len(shape)
    )
    if not c0:
        _err_ind(ind_ch, shape, key)

    # check all parts are np.ndarrays
    ind_ch = tuple([
        ii if isinstance(ii, (slice, np.ndarray)) else np.r_[ii]
        for ii in ind_ch
    ])

    # ------------------------
    # test
    # ------------------------

    try:
        aa = np.empty(shape)[ind_ch]
        del aa
    except Exception as err:
        _err_ind(ind_ch, shape, key, err=err)

    return ind_ch


def _err_ind(ind_ch, shape, key, err=None):
    msg = (
        f"Arg ind_ch must be compatible with shape of rays '{key}':\n"
        f"\t- rays shape: {shape}\n"
        f"\t- ind_ch: {ind_ch}\n"
    )
    if err is not None:
        msg += "\n\n{err}"

    raise Exception(msg)


# ###############################################################
#                   remove consecutive nans
# ###############################################################


def remove_consecutive_nans(arr):

    # -----------------
    # preliminary check
    # -----------------

    assert arr.ndim == 1, f"arr must be 1d: {arr.shape}"

    # -------------------
    # get indices on nans
    # -------------------

    ind = np.nonzero(np.isnan(arr))[0]
    cons = np.r_[False, np.diff(ind) == 1]

    return np.delete(arr, ind[cons])